import copy
import itertools
import numpy as np
from scipy.sparse import dok_matrix

import pyomo.environ as pe
from pyomo.repn import generate_standard_repn
from pyomo.core.base import SortComponents, is_fixed

from pao.mpr import LinearMultilevelProblem, QuadraticMultilevelProblem
from .components import SubModel


def _bound_equals(exp, num):
    if exp is None:
        return False
    if is_fixed(exp):
        return pe.value(exp) == num
    return False                        # pragma: no cover
                                        # WEH - will we ever reach this point?  What if the LB is a mutable parameter?

def offset(t,L):
    if t==0:
        return 0
    elif t==1:
        return L.nxR
    else:
        return L.nxR+L.nxZ

class Node(object):

    global_list = []

    def __init__(self, node):
        self.nid = len(Node.global_list)
        Node.global_list.append(self)
        self.node = node
        self.children = []
        self.orepn = []
        self.crepn = []
        self.fixedvars = set()
        self.unfixedvars = set()    # unfixed variables used in expressions
        self.xR = {}
        self.xZ = {}
        self.xB = {}
        self.linear = True
        #
        # Collect vardata that are declared fixed.  If a variable component
        # is specified, then collect all of its vardata objects.
        #
        for v in getattr(node, '_fixed', []):
            if v.is_indexed():
                for vardata in v.values():
                    self.fixedvars.add(id(vardata))
            else:
                self.fixedvars.add(id(v))

    def is_linear(self):
        if not self.linear:
            return False
        for child in self.children:
            if not child.linear:
                return False
        return True 

    def categorize_variable(self, vid, v, vidmap):
        if v.is_binary():
            vidmap[vid] = (2, self.nid, len(self.xB))
            self.xB[len(self.xB)] = vid

        elif v.is_integer():
            #
            # The _bounds_equals function only returns True if the 
            # bound is a fixed value and its value equals the 2nd argument.
            # If the bound value is specified with Pyomo Parameters and their
            # value equals, then we are assuming the multilevel problem will
            # be re-generated if those parameter values change.
            #
            if _bound_equals(v.lb, 0) and _bound_equals(v.ub, 1):
                vidmap[vid] = (2, self.nid, len(self.xB))
                self.xB[len(self.xB)] = vid
            else:
                vidmap[vid] = (1, self.nid, len(self.xZ))
                self.xZ[len(self.xZ)] = vid

        else:
            assert (v.is_continuous()), "Variable '%s' has a domain type that is not continuous, integer or binary" % str(v)
            vidmap[vid] = (0, self.nid, len(self.xR))
            self.xR[len(self.xR)] = vid

    def add_levels(self, L, levelmap, treemap):
        treemap[self.nid] = self
        levelmap[L.id] = L
        for child in self.children:
            L_ = L.add_lower(id=child.nid)
            child.add_levels(L_, levelmap, treemap)

    def initialize_level_vars(self, level, inequalities, var):
        """
        Initialize the level object...
        """
        level.x._resize(nxR=len(self.xR), nxZ=len(self.xZ), nxB=len(self.xB))
        #
        # xR
        #
        for i in self.xR:
            vid = self.xR[i]
            val = var[vid].lb
            if val is not None:
                level.x.lower_bounds[i] = val
            val = var[vid].ub
            if val is not None:
                level.x.upper_bounds[i] = val
        #
        # xZ
        #
        for i in self.xZ:
            vid = self.xZ[i]
            val = var[vid].lb
            if val is not None:
                level.x.lower_bounds[i+level.x.nxR] = val
            val = var[vid].ub
            if val is not None:
                level.x.upper_bounds[i+level.x.nxR] = val
        #
        # xB
        #
        for i in self.xB:
            level.x.lower_bounds[i+level.x.nxR+level.x.nxZ] = 0
            level.x.upper_bounds[i+level.x.nxR+level.x.nxZ] = 1

    def initialize_level(self, level, inequalities, var, vidmap, levelmap):
        #
        # Objective
        #
        assert (len(self.orepn) <= 1), "PAO model has %d objectives specified, but a MultilevelProblem can have no more than one" % len(self.orepn)
        if len(self.orepn) == 1:
            repn = self.orepn[0][0]
            #
            # minimize
            #
            if self.orepn[0][1] == pe.maximize:
                level.minimize = False
            #
            # d
            #
            level.d = pe.value(repn.constant)
            #
            # c
            #
            c = {}
            for j in levelmap:
                node = levelmap[j]
                c[j] = [0]*node.x.num

            for i,val in enumerate(repn.linear_coefs):
                vid = id(repn.linear_vars[i])
                t, nid, j = vidmap[vid]
                L = levelmap[nid]
                c[nid][j+offset(t,L.x)] = pe.value(val)

            # Add a non-null objective vector
            for j in levelmap:
                for v in c[j]:
                    if v != 0:
                        level.c[j] = c[j]
                        break
            #
            # P
            #
            P = {}
            for i,val in enumerate(repn.quadratic_coefs):
                v1,v2 = repn.quadratic_vars[i]
                t1, nid1, j1 = vidmap[id(v1)]
                t2, nid2, j2 = vidmap[id(v2)]
                L1 = levelmap[nid1]
                L2 = levelmap[nid2]
                if nid1 <= nid2:
                    if P.get((nid1,nid2),None) is None:
                        P[nid1,nid2] = {}
                    P[nid1,nid2][j1+offset(t1,L1.x), j2+offset(t2,L2.x)] = pe.value(val)
                else:
                    if P.get((nid2,nid1),None) is None:
                        P[nid2,nid1] = {}
                    P[nid2,nid1][j2+offset(t2,L2.x), j1+offset(t1,L1.x)] = pe.value(val)
            for n1,n2 in P:
                level.P[n1,n2] = (len(levelmap[n1].x),len(levelmap[n2].x)), P[n1,n2]
        #
        # Constraints
        #
        if len(self.crepn) > 0:
            #
            # A
            #
            A = {}
            for i in levelmap:
                A[i] = {}
            nrows = len(self.crepn)

            for k in range(len(self.crepn)):
                repn = self.crepn[k][0]
                for i,c  in enumerate(repn.linear_coefs):
                    vid = id(repn.linear_vars[i])
                    t, nid, j = vidmap[vid]
                    L = levelmap[nid]
                    c_ = pe.value(c)
                    if c_ != 0:
                        A[nid][k,j+offset(t,L.x)] = c_

            for j in levelmap:
                if len(A[j]) > 0:
                    L = levelmap[j]
                    mat = dok_matrix((nrows, L.x.num))
                    for key, val in A[j].items():
                        mat[key] = val
                    level.A[L.id] = mat
            #
            # Q
            #
            Q = {}
            for k in range(len(self.crepn)):
                repn = self.crepn[k][0]
                for i,val in enumerate(repn.quadratic_coefs):
                    v1,v2 = repn.quadratic_vars[i]
                    t1, nid1, j1 = vidmap[id(v1)]
                    t2, nid2, j2 = vidmap[id(v2)]
                    L1 = levelmap[nid1]
                    L2 = levelmap[nid2]
                    if nid1 <= nid2:
                        if Q.get((nid1,nid2),None) is None:
                            Q[nid1,nid2] = {}
                        Q[nid1,nid2][k, j1+offset(t1,L1.x), j2+offset(t2,L2.x)] = pe.value(val)
                    else:
                        if Q.get((nid2,nid1),None) is None:
                            Q[nid2,nid1] = {}
                        Q[nid2,nid1][k, j2+offset(t2,L2.x), j1+offset(t1,L1.x)] = pe.value(val)
            for n1,n2 in Q:
                level.Q[n1,n2] = (len(self.crepn), len(levelmap[n1].x),len(levelmap[n2].x)), Q[n1,n2]
            #
            # b
            #
            b = []
            for k in range(len(self.crepn)):
                repn = self.crepn[k][0]
                b.append(self.crepn[k][1] - repn.constant)
            level.b = b


def negate_repn(repn):
    trepn = copy.copy(repn)
    trepn.constant *= -1
    trepn.linear_coefs = list(-1*repn.linear_coefs[i] for i in range(len(repn.linear_coefs)))
    trepn.linear_vars = list(repn.linear_vars)
    trepn.quadratic_coefs = list(-1*repn.quadratic_coefs[i] for i in range(len(repn.quadratic_coefs)))
    trepn.quadratic_vars = list(repn.quadratic_vars)
    return trepn


def collect_multilevel_tree(block, var, vidmap={}, sortOrder=SortComponents.unsorted, fixed=set(), inequalities=None):
    """
    Traverse the model and generate a tree of the SubModel components
    """
    #
    # Roof of the current subtree, defined by the block
    #
    curr = Node(block)
    #
    # Recurse, collecting Submodel components
    #
    fixedvars = fixed | curr.fixedvars
    curr.children = \
        [collect_multilevel_tree(submodel, var, vidmap, fixed=fixedvars, inequalities=inequalities) 
         for submodel in block.component_objects(SubModel, active=True, descend_into=True, sort=sortOrder)]
    #
    # Collect objectives and constraints in the current submodel.
    # Note that we do not recurse into SubModel blocks.
    #
    # Objectives
    #
    for odata in block.component_data_objects(pe.Objective, active=True, sort=sortOrder, descend_into=True):
        repn = generate_standard_repn(odata.expr)
        degree = repn.polynomial_degree()
        assert (degree is not None), "Objective '%s' has a body that is not linear or quadratic" % odata.name
        if degree == 0:
            continue # trivial, so skip
        curr.orepn.append( (repn, odata.sense) )
        if degree == 2:
            curr.linear = False
    #
    # Constraints
    #
    # If we call conversion twice, then we delete the variables from the previous conversion
    #
    block.del_component('zzz_PAO_SlackVariables')
    block.del_component('zzz_PAO_SlackVariables_index')
    block.zzz_PAO_SlackVariables = pe.VarList(domain=pe.NonNegativeReals)
    for cdata in block.component_data_objects(pe.Constraint, active=True, sort=sortOrder, descend_into=True):
        if (not cdata.has_lb()) and (not cdata.has_ub()):
            assert not cdata.equality, "Constraint '%s' is an equality with an infinite right-hand-side" % cdata.name
            # non-binding, so skip
            continue                            # pragma: no cover
        repn = generate_standard_repn(cdata.body)
        degree = repn.polynomial_degree()
        assert (degree is not None), "Constraint '%s' has a body that is not linear or quadratic " % cdata.name
        if degree == 0:
            if cdata.equality:
                assert pe.value(cdata.body) == pe.value(cdata.lower), "Constraint '%s' is constant but it is not satisfied (equality)" % cdata.name
            else:
                if not cdata.lower is None:
                    assert pe.value(cdata.body) >= pe.value(cdata.lower), "Constraint '%s' is constant but it is not satisfied (lower-bound)" % cdata.name
                if not cdata.upper is None:
                    assert pe.value(cdata.body) <= pe.value(cdata.upper), "Constraint '%s' is constant but it is not satisfied (upper-bound)" % cdata.name
            # trivial, so skip
            continue                            # pragma: no cover
        else:
            if degree == 2:
                curr.linear = False
            if inequalities:
                if cdata.equality:
                    val = pe.value(cdata.lower)
                    curr.crepn.append( (repn, val) )
                    curr.crepn.append( (negate_repn(repn), -val) )
                else:
                    if cdata.lower is None and cdata.upper is None:             #pragma: no cover
                        # unbounded constraint
                        continue
                    if cdata.lower is not None:
                        curr.crepn.append( (negate_repn(repn), -pe.value(cdata.lower)) )
                    if cdata.upper is not None:
                        curr.crepn.append( (repn, pe.value(cdata.upper)) )
            else:
                if cdata.equality:
                    curr.crepn.append( (repn, pe.value(cdata.lower)) )
                else:
                    if cdata.lower is None and cdata.upper is None:             #pragma: no cover
                        # unbounded constraint
                        continue
                    if cdata.lower is not None:
                        trepn = negate_repn(repn)
                        trepn.linear_coefs.append(1)
                        trepn.linear_vars.append(block.zzz_PAO_SlackVariables.add())
                        curr.crepn.append( (trepn, -pe.value(cdata.lower)) )
                    if cdata.upper is not None:
                        repn.linear_coefs = list(repn.linear_coefs)
                        repn.linear_vars = list(repn.linear_vars)
                        repn.linear_coefs.append(1)
                        repn.linear_vars.append(block.zzz_PAO_SlackVariables.add())
                        curr.crepn.append( (repn, pe.value(cdata.upper)) )
    #
    # Collect the variables used by the children
    #
    childvars = set()
    for child in curr.children:
        childvars |= child.unfixedvars
    #
    # Add ids for variables in this block that have not been specified 
    # as fixed and which are not used in submodels
    #
    knownvars = curr.fixedvars | childvars
    newvars = []
    for repn in itertools.chain(curr.orepn, curr.crepn):
        for v in repn[0].linear_vars:
            i = id(v)
            if i not in knownvars:
                curr.unfixedvars.add(i)
                var[i] = v
                newvars.append(i)
                knownvars.add(i)
        if True:                                   # pragma: no cover
            for v,w in repn[0].quadratic_vars:
                i = id(v)
                if i not in knownvars:
                    curr.unfixedvars.add(i)
                    var[i] = v
                    newvars.append(i)
                    knownvars.add(i)
                i = id(w)
                if i not in knownvars:
                    curr.unfixedvars.add(i)
                    var[i] = w
                    newvars.append(i)
                    knownvars.add(i)
    #
    # Categorize the new variables that were found
    #
    if sortOrder == SortComponents.unsorted:
        for i in newvars:
            curr.categorize_variable(i,var[i], vidmap)
    else:
        for k,_,w in sorted(((i,var[i].name,var[i]) for i in newvars), key=lambda arg: arg[1]):
                curr.categorize_variable(k,w, vidmap)
    #
    # Return root of this tree
    #
    return curr


class PyomoSubmodel_SolutionManager_LBP(object):

    def __init__(self, var, vidmap, pyomo_id):
        self.var = var
        self.vidmap = vidmap
        self.pyomo_id = pyomo_id

    def copy(self, *, From, To):
        assert (id(To) == self.pyomo_id), "Attempting to copy data into a different model than was used to create the multilevel problem"
        for vid in self.vidmap:
            v = self.var[vid]
            t, nid, j = self.vidmap[vid]
            if nid == 0:
                U = From.U
                v.value = From.U.x.values[j+offset(t,U.x)]
            else:
                L = From.U.LL[nid-1]
                v.value = From.U.LL[nid-1].x.values[j+offset(t,L.x)]


def convert_pyomo2LinearMultilevelProblem(model, *, determinism=1, inequalities=True):
    """
    Traverse the model and generate a LinearMultilevelProblem.  Generate errors
    if this problem cannot be represented in this form.

    Args
    ---- 
    model
        A Pyomo model object.
    determinism: int
        Indicates whether the traversal of **model** is
        ordered.  Valid values are:

                * 0 - Unordered traversal of **model**
                * 1 - Ordered traversal of component indices in **model**
                * 2 - Ordered traversal of components by name in **model**

    inequalities: bool
        If True, then the LinearMultilevelProblem object represents all
        constraints as less-than-or-equal inequalities.  Otherwise,
        the LinearMultilevelProblem represents all constraints as equalities.

    Returns
    -------
    LinearMultilevelProblem
        This object corresponds to the problem in **model**.
    """
    return convert_pyomo2MultilevelProblem(model, determinism=determinism, inequalities=inequalities, linear=True)


def convert_pyomo2QuadraticMultilevelProblem(model, *, determinism=1, inequalities=True):
    """
    Traverse the model an generate a QuadraticMultilevelProblem.  Generate errors
    if this problem cannot be represented in this form.

    Args
    ---- 
    model
        A Pyomo model object.
    determinism: int
        Indicates whether the traversal of **model** is
        ordered.  Valid values are:

                * 0 - Unordered traversal of **model**
                * 1 - Ordered traversal of component indices in **model**
                * 2 - Ordered traversal of components by name in **model**

    inequalities: bool
        If True, then the QuadraticMultilevelProblem object represents all
        constraints as less-than-or-equal inequalities.  Otherwise,
        the QuadraticMultilevelProblem represents all constraints as equalities.

    Returns
    -------
    QuadraticMultilevelProblem
        This object corresponds to the problem in **model**.
    """
    return convert_pyomo2MultilevelProblem(model, determinism=determinism, inequalities=inequalities, linear=False)


def convert_pyomo2MultilevelProblem(model, *, determinism=1, inequalities=True, linear=None):
    """
    Traverse the model an generate a LinearMultilevelProblem or
    QuadraticMultilevelProblem.  Generate errors if this problem cannot
    be represented in this form.

    Args
    ---- 
    model
        A Pyomo model object.
    determinism: int, Default: 1
        Indicates whether the traversal of **model** is
        ordered.  Valid values are:

                * 0 - Unordered traversal of **model**
                * 1 - Ordered traversal of component indices in **model**
                * 2 - Ordered traversal of components by name in **model**

    linear: bool
        A flag that indicates whether the expected model representation is linear (True) or 
        quadratic (False).  If not specified, then no error checking is done to confirm
        whether the model is linear or quadratic.
    inequalities: bool, Default: True
        If True, then the multilevel problem object represents all
        constraints as less-than-or-equal inequalities.  Otherwise,
        the multilevel problem represents all constraints as equalities.

    Returns
    -------
    LinearMultilevelProblem or QuadraticMultilevelProblem
        This object corresponds to the problem in **model**.
    """
    #
    # Cleanup global memory
    #
    Node.global_list = []
    #
    # Define sort order for searching Pyomo model
    #
    sortOrder = SortComponents.unsorted
    if determinism >= 1:
        sortOrder = sortOrder | SortComponents.indices
        if determinism >= 2:
            sortOrder = sortOrder | SortComponents.alphabetical
    #
    # Collect tree of models
    #
    var = {}
    vidmap = {}
    tree = collect_multilevel_tree(model, var, vidmap, sortOrder=sortOrder, inequalities=inequalities)
    #
    # We must have a least one SubModel
    #
    assert (len(tree.children) > 0), "Pyomo problem does not contain SubModel components"
    #
    # Collect SubModel representations
    #
    treemap = {}
    levelmap = {}
    if linear is None:
        linear = tree.is_linear()
    if linear is True:
        M = LinearMultilevelProblem()
    else:
        M = QuadraticMultilevelProblem()
    U = M.add_upper(id=tree.nid)
    treemap[tree.nid] = tree

    tree.add_levels(U, levelmap, treemap)
    
    for L in M.levels():
        node = treemap[L.id]
        node.initialize_level_vars(L, inequalities, var)
    
    for L in M.levels():
        node = treemap[L.id]
        node.initialize_level(L, inequalities, var, vidmap, levelmap)
    #
    # Cleanup global memory
    #
    Node.global_list = []

    return M, PyomoSubmodel_SolutionManager_LBP(var, vidmap, id(model))
    

convert_pyomo2lmp = convert_pyomo2LinearMultilevelProblem
convert_pyomo2qmp = convert_pyomo2QuadraticMultilevelProblem
convert_pyomo2mp = convert_pyomo2MultilevelProblem




# pao.mpr.examples

from . import anadalingam
from . import bard511_list
from . import bard511
from . import barguel
from . import besancon27
from . import besancon27_shifted
from . import getachew_ex1
from . import getachew_ex2
from . import pineda
from . import toyexample1
from . import toyexample2
from . import toyexample3



import math
import pyutilib.th as unittest
import pyomo.environ as pe

from pao import Solver
import pyomo.opt
from pyomo.neos.kestrel import kestrelAMPL

neos_available = False
try:
    raise RuntimeError("Disable NEOS Tests")
    if kestrelAMPL().neos is not None:
        neos_available = True
except:
    pass

solvers = pyomo.opt.check_available_solvers('glpk','ipopt')


def create_lp1():
    M = pe.ConcreteModel()
    M.x = pe.Var(bounds=(0,None))
    M.y = pe.Var(bounds=(0,None))
    M.o = pe.Objective(expr=2*M.x+M.y, sense=pe.maximize)
    M.c = pe.Constraint(expr=M.x+M.y == 1)
    return M

def create_nlp1():
    M = pe.ConcreteModel()
    A = list(range(10))
    M.x = pe.Var(A, bounds=(0,None), initialize=1)
    M.o = pe.Objective(expr=sum(pe.sin((i+1)*M.x[i]) for i in A))
    M.c = pe.Constraint(expr=sum(M.x[i] for i in A) >= 1)
    return M


@unittest.skipIf('ipopt' not in solvers, "Ipopt solver is not available")
class Test_ipopt(unittest.TestCase):

    def test_lp1(self):
        M = create_lp1()
        opt = Solver('ipopt')
        self.assertTrue(opt.available())
        res = opt.solve(M)

        self.assertTrue(math.isclose(M.x.value, 1, abs_tol=1e-6))
        self.assertTrue(math.isclose(M.y.value, 0, abs_tol=1e-6))

    def test_nlp1(self):
        M = create_nlp1()
        opt = Solver('ipopt', tee=True, max_cpu_time=1e-12, print_level=0)
        self.assertTrue(opt.available())
        #opt.config.display()
        #print(opt.solver.options)
        res = opt.solve(M)
        res = opt.solve(M, max_cpu_time=100)
        # Solving with a solver option doesn't change the value configured when creating the solver
        self.assertTrue(math.isclose(opt.solver.options.max_cpu_time, 1e-12))

@unittest.skipIf('glpk' not in solvers, "GLPK solver is not available")
class Test_glpk(unittest.TestCase):

    def test_lp1(self):
        M = create_lp1()
        opt = Solver('glpk')
        self.assertTrue(opt.available())
        res = opt.solve(M)

        self.assertTrue(math.isclose(M.x.value, 1, abs_tol=1e-6))
        self.assertTrue(math.isclose(M.y.value, 0, abs_tol=1e-6))


class Test_foobar(unittest.TestCase):

    def test_lp1(self):
        try:
            Solver('foobar')
        except AssertionError:
            pass

    def test_glpk_executable(self):
        try:
            opt = Solver('glpk', executable="glpk_foobar")
        except ValueError:
            pass
        #
        opt = Solver('glpk')
        opt.config.executable = 'glpk_foobar'
        try:
            self.assertTrue(opt.available())
        except ValueError:
            pass
        #
        M = create_lp1()
        opt = Solver('glpk')
        opt.config.executable = 'glpk_foobar'
        try:
            self.assertTrue(opt.solve(M))
        except ValueError:
            pass



class Test_neos(unittest.TestCase):

    @unittest.skipIf(not neos_available, "NEOS not available")
    def test_neos_ipopt_available(self):
        M = create_nlp1()
        opt = Solver('ipopt', server='neos', email='pao@gmail.com', max_cpu_time=1e-12)
        self.assertTrue(opt.available())
        res = opt.solve(M)
        #M.x.pprint()
        res = opt.solve(M, max_cpu_time=100)
        self.assertTrue(math.isclose(opt.solver_options['max_cpu_time'], 1e-12))
        #print(res)
        #M.x.pprint()

    @unittest.skipIf(not neos_available, "NEOS not available")
    def test_neos_cbc_available(self):
        M = create_lp1()
        opt = Solver('cbc', server='neos', email='pao@gmail.com')
        self.assertTrue(opt.available())
        res = opt.solve(M, tee=True)
        
        self.assertTrue(math.isclose(M.x.value, 1, abs_tol=1e-6))
        self.assertTrue(math.isclose(M.y.value, 0, abs_tol=1e-6))

    @unittest.skipIf(not neos_available, "NEOS not available")
    def test_neos_cbc_options(self):
        M = create_lp1()
        opt = Solver('cbc', server='neos', email='pao@gmail.com', foo=1, bar=None)
        self.assertTrue(opt.available())
        res = opt.solve(M)
        
        self.assertTrue(math.isclose(M.x.value, 1, abs_tol=1e-6))
        self.assertTrue(math.isclose(M.y.value, 0, abs_tol=1e-6))

    # This is slow
    @unittest.skipIf(neos_available, "NEOS is available")
    def Xtest_neos_cbc_missing(self):
        try:
            opt = Solver('cbc', server='neos')
        except AssertionError:
            pass

    @unittest.skipIf(not neos_available, "NEOS not available")
    def test_neos_foobar(self):
        M = create_lp1()
        opt = Solver('foobar', server='neos')
        try:
            res = opt.solve(M)
        except RuntimeError:
            pass


if __name__ == "__main__":
    unittest.main()



#
# Trilevel example 
#
# Anandalingam, G.: A mathematical programming model of decentralized multi-level systems. 
# J. Oper.Res. Soc.39(11), 1021–1033 (1988)
#

import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()
    M.x1 = pe.Var(bounds=(0,None))
    M.x2 = pe.Var(bounds=(0,None))
    M.x3 = pe.Var(bounds=(0,0.5))

    M.L = SubModel(fixed=M.x1)

    M.L.B = SubModel(fixed=M.x2)

    M.o = pe.Objective(expr=-7*M.x1 - 3*M.x2 + 4*M.x3)

    M.L.o = pe.Objective(expr=-M.x2)
    M.L.B.o = pe.Objective(expr=-M.x3)

    M.L.B.c1 = pe.Constraint(expr=   M.x1 + M.x2 + M.x3 <= 3)
    M.L.B.c2 = pe.Constraint(expr=   M.x1 + M.x2 - M.x3 <= 1)
    M.L.B.c3 = pe.Constraint(expr=   M.x1 + M.x2 + M.x3 >= 1)
    M.L.B.c4 = pe.Constraint(expr= - M.x1 + M.x2 + M.x3 <= 1)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.pprint()



#
# Simple interdiction problem solvable through
# lower level dualization and MPEC transformation
#
# Note that this is a quadratic bilevel problem
#
# This model has an optimal objective value of 0
# To see this, fix u and optimize the lower level
#
import pyomo.environ as pe
from pao.pyomo import SubModel


def create():
    M = pe.ConcreteModel()
    M.u = pe.Var(within=pe.Binary, initialize=1)
    M.x = pe.Var(bounds=(0,1)) # dual var for UB: v1
    M.y = pe.Var(bounds=(0,1)) # dual var for UB: v2

    # Note that the upper and lower level objectives
    # are the same.
    M.o = pe.Objective(expr=M.x + M.u)

    M.sub = SubModel(fixed=M.u)
    M.sub.o = pe.Objective(expr=M.x + M.u, sense=pe.maximize)
    M.sub.c = pe.Constraint(expr= M.x == M.y*M.u) # x - y*u = 0, dual var: u1

    return M


######################
# DUALIZED:
# min   v1 + v2 + u
# s.t.  u1 + v1 >= 1    : x
#       -u*u1 + v2 >= 0  : y
#       v1, v2, >= 0
#       u1 unconstrained



#
# Toy Example 3 from 
#
# "A projection-based reformulation and decomposition algorithm for global optimization 
#  of a class of mixed integer bilevel linear programs" by Dajun Yue, Jiyao Gao, Bo Zeng, Fengqi You
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxR=1, nxZ=1)
    U.x.lower_bounds = [0,0]

    L = U.add_lower(nxR=1, nxZ=1)
    L.x.lower_bounds = [0, 0]

    U.minimize = True
    U.c[U] = [20, -38]
    U.c[L] = [1, 42]

    U.A[U] = [[7, 5],
              [6, 9]]
    U.A[L] = [[0, 7],
              [10, 2]]
    U.b = [62, 117]

    L.maximize = True
    L.c[L] = [39, 27]

    L.A[U] = [[8, 0],
              [9, 0]]
    L.A[L] = [[2, 8], 
              [2, 1]]
    L.b = [53, 28]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.PCCG')
    opt.solve(M)
    M.print()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao

This package defines modeling techniques and solvers for representing
optimization problems with adversarial behavior.  PAO currently supports
bilevel programming.

Importing pao initializes the Pyomo environment and then registers
the pao plugins.  We assume that a user will never import symbols from
pao directly:

    $ from pao import *

Instead, users should import symbols directly from pao sub-packages:

    $ from pao.pyomo import *

Version: %s
"""

__version__ = '1.0.dev0'
__doc__ = __doc__ % __version__

__all__ = ('__version__')

import pyomo.environ
import pao.mpr
import pao.pyomo
#import pao.duality.plugins
import pao.pyomo.plugins
pao.pyomo.plugins.load()

from pao.common import Solver



# 
# MIBS generalExample
#
# https://coral.ise.lehigh.edu/wp-content/uploads/2016/02/MibS_inputFile.pdf
#
# The optimal solution is C0=6, C1=5
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.x = pe.Var(bounds=(0,10), within=pe.Integers)
    M.y = pe.Var(bounds=(0,5), within=pe.Integers)

    M.o = pe.Objective(expr=-M.x - 7*M.y, sense=pe.minimize)
    M.c1 = pe.Constraint(expr= -3*M.x + 2*M.y <= 12)
    M.c2 = pe.Constraint(expr=    M.x + 2*M.y <= 20)

    M.L = SubModel(fixed=M.x)
    M.L.o = pe.Objective(expr=M.y, sense=pe.minimize)
    M.L.c1 = pe.Constraint(expr=  2*M.x -   M.y <= 7)
    M.L.c2 = pe.Constraint(expr= -2*M.x + 4*M.y <= 16)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    #M.pprint()
    opt = Solver("pao.pyomo.MIBS")
    results = opt.solve(M, tee=True)
    print(pe.value(M.x))
    print(pe.value(M.y))



#
# Utilities for creating Pyomo models using data in
# LinearMultilevelProblem objects.
#
import numpy as np
import pyomo.environ as pe
import pyomo.opt.results
from ..repn import LinearLevelRepn, LevelValues, SimplifiedList, LevelVariable
import pao.common.solver


def dot(A, x, num=None):
    if A is None:
        if num is not None:
            if num > 1:
                return np.zeros(num)
            return 0
        return None
    #if type(x) is SimplifiedList:
    #    x = x[0]
    if type(x) is LevelVariable:
        x = x.pyvar
    #if type(A) is SimplifiedList:
    #    A = A[0]

    if type(A) is np.ndarray:
        return sum(A*x)
    else:
        Acoo = A.tocoo()
        e = [0] * Acoo.shape[0]
        for i,j,v in zip(Acoo.row, Acoo.col, Acoo.data):
            e[i] += v*x[j]
        return np.array(e)

def add_variables(block, level):
    pyvar = []
    if level.x.nxR > 0:
        block.xR = pe.Var(range(0,level.x.nxR), within=pe.Reals)
        for i in range(level.x.nxR):
            pyvar.append(block.xR[i])
    else:
        block.xR = None
    if level.x.nxZ > 0:
        block.xZ = pe.Var(range(0,level.x.nxZ), within=pe.Integers)
        for i in range(level.x.nxZ):
            pyvar.append(block.xZ[i])
    else:
        block.xZ = None
    if level.x.nxB > 0:
        block.xB = pe.Var(range(0,level.x.nxB), within=pe.Binary)
        for i in range(level.x.nxB):
            pyvar.append(block.xB[i])
    else:
        block.xB = None

    level.x.pyvar = np.array(pyvar)

    if level.x.lower_bounds is not None:
        i=0
        while i<level.x.nxR:
            lb = level.x.lower_bounds[i]
            if not lb is np.NINF:
                block.xR[i].setlb( lb )
            i += 1
        while i<level.x.nxZ:
            lb = level.x.lower_bounds[i]
            if not lb is np.NINF:
                block.xZ[i].setlb( lb )
            i += 1
        while i<level.x.nxB:
            lb = level.x.lower_bounds[i]
            block.xB[i].setlb( lb )
            i += 1

    if level.x.upper_bounds is not None:
        i=0
        while i<level.x.nxR:
            ub = level.x.upper_bounds[i]
            if not ub is np.PINF:
                block.xR[i].setub( ub )
            i += 1
        while i<level.x.nxZ:
            ub = level.x.upper_bounds[i]
            if not ub is np.PINF:
                block.xZ[i].setub( ub )
            i += 1
        while i<level.x.nxB:
            ub = level.x.upper_bounds[i]
            block.xB[i].setub( ub )
            i += 1


def add_linear_constraints(block, A, U, L, b, inequalities):
    assert (b is not None), "Unexpected 'None' value for constraint RHS"
    nc = b.size
    if nc == 0:
        return
    e = dot(A[U], U.x, num=nc) + dot(A[L], L.x, num=nc)

    block.c = pe.ConstraintList()
    for i in range(len(e)):
        if type(e[i]) in [int,float]:
            if inequalities:
                assert e[i] <= b[i], "Trivial linear constraint violated: %f <= %f" % (e[i], b[i])
            else:
                assert e[i] == b[i], "Trivial linear constraint violated: %f == %f" % (e[i], b[i])
            continue
        if inequalities:
            block.c.add( e[i] <= b[i] )
        else:
            block.c.add( e[i] == b[i] )


def pyomo2pao_termination_condition(tc):

    if tc == pyomo.opt.results.TerminationCondition.unknown or \
       tc == pyomo.opt.results.TerminationCondition.other or \
       tc == pyomo.opt.results.TerminationCondition.feasible or  \
       tc == pyomo.opt.results.TerminationCondition.maxEvaluations:
        return pao.common.solver.TerminationCondition.unknown

    elif tc == pyomo.opt.results.TerminationCondition.maxTimeLimit:
        return pao.common.solver.TerminationCondition.maxTimeLimit

    elif tc == pyomo.opt.results.TerminationCondition.maxIterations:
        return pao.common.solver.TerminationCondition.maxIterations

    elif tc == pyomo.opt.results.TerminationCondition.minFunctionValue:
        return pao.common.solver.TerminationCondition.objectiveLimit

    elif tc == pyomo.opt.results.TerminationCondition.minStepLength:
        return pao.common.solver.TerminationCondition.minStepLength

    elif tc == pyomo.opt.results.TerminationCondition.optimal or \
         tc == pyomo.opt.results.TerminationCondition.globallyOptimal or \
         tc == pyomo.opt.results.TerminationCondition.locallyOptimal:
        return pao.common.solver.TerminationCondition.optimal

    elif tc == pyomo.opt.results.TerminationCondition.unbounded:
        return pao.common.solver.TerminationCondition.unbounded

    elif tc == pyomo.opt.results.TerminationCondition.infeasible:
        return pao.common.solver.TerminationCondition.infeasible

    elif tc == pyomo.opt.results.TerminationCondition.infeasibleOrUnbounded:
        return pao.common.solver.TerminationCondition.infeasibleOrUnbounded

    elif tc == pyomo.opt.results.TerminationCondition.invalidProblem or \
         tc == pyomo.opt.results.TerminationCondition.noSolution or \
         tc == pyomo.opt.results.TerminationCondition.internalSolverError or \
         tc == pyomo.opt.results.TerminationCondition.solverFailure or \
         tc == pyomo.opt.results.TerminationCondition.intermediateNonInteger:
        return pao.common.solver.TerminationCondition.error

    elif tc == pyomo.opt.results.TerminationCondition.userInterrupt or \
         tc == pyomo.opt.results.TerminationCondition.resourceInterrupt:
        return pao.common.solver.TerminationCondition.interrupted

    elif tc == pyomo.opt.results.TerminationCondition.licensingProblems:
        return pao.common.solver.TerminationCondition.licensingProblems 

    raise RuntimeError("Unknown Pyomo termination condition: "+str(tc))
    



#
# Trilevel example 
#
# Anandalingam, G.: A mathematical programming model of decentralized multi-level systems. 
# J. Oper.Res. Soc.39(11), 1021–1033 (1988)
#
# This example illustrates how pao.mpr can model tri-level problems, though
# we don't have PAO solvers that support these yet.
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxR=1)
    U.x.lower_bounds = [0]

    L = U.add_lower(nxR=1)
    L.x.lower_bounds = [0]

    B = L.add_lower(nxR=1)
    B.x.lower_bounds = [0]
    B.x.upper_bounds = [0.5]

    U.minimize = True
    U.c[U] = [-7]
    U.c[L] = [-3]
    U.c[B] = [4]

    L.minimize = True
    L.c[L] = [-1]

    B.minimize = True
    B.c[B] = [-1]

    B.inequalities = True
    B.A[U] = [[1], [ 1], [-1], [-1]]
    B.A[L] = [[1], [ 1], [-1], [ 1]]
    B.A[B] = [[1], [-1], [-1], [ 1]]
    B.b = [3,1,-1,1]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.PCCG')
    opt.solve(M)
    M.print()



#
# Solvers that convert the problem to a LinearBilevelProblem and
# solve using a LBP-specific solver.
#
import time
import pao.mpr
import pao.common
from pao.pyomo.solver import Solver, PyomoSubmodelSolverBase_LBP
from pyomo.common.config import ConfigBlock, ConfigValue


@Solver.register(
        name='pao.pyomo.FA',
        doc="PAO solver for Pyomo models that define linear and bilinear bilevel problems.  Solver uses big-M relaxations discussed by Fortuny-Amat and McCarl (1981).")
class PyomoSubmodelSolver_FA(PyomoSubmodelSolverBase_LBP):
    """
    PAO FA solver for Pyomo models: pao.pyomo.FA

    This solver converts the Pyomo model to a LinearBilevelProblem and
    calls the pao.mpr.FA solver.
    """

    config = PyomoSubmodelSolverBase_LBP.config()
    config.declare('mip_solver', ConfigValue(
        default='glpk',
        description="The MIP solver used by FA.  (default is glpk)"
        ))
    #config.declare('mip_options', ConfigValue(
    #    default=None,
    #    description="A dictionary that defines the solver options for the MIP solver.  (default is None)"))
    
    def __init__(self, **kwds):
        super().__init__('pao.pyomo.FA', 'pao.mpr.FA')

    def solve(self, model, **options):
        return super().solve(model, **options)

pao.common.SolverAPI._generate_solve_docstring(PyomoSubmodelSolver_FA)



@Solver.register(
        name='pao.pyomo.MIBS',
        doc="PAO solver for Multilevel Problem Representations using the COIN-OR MibS solver by Tahernejad, Ralphs, and DeNegre (2020).")
class PyomoSubmodelSolver_MIBS(PyomoSubmodelSolverBase_LBP):
    """
    PAO MibS solver for Pyomo models: pao.pyomo.MIBS

    This solver converts the Pyomo model to a LinearBilevelProblem and
    calls the pao.mpr.MIBS solver.
    """

    config = PyomoSubmodelSolverBase_LBP.config()
    
    def __init__(self, **kwds):
        super().__init__('pao.pyomo.MIBS', 'pao.mpr.MIBS')

    def solve(self, model, **options):
        return super().solve(model, **options)

pao.common.SolverAPI._generate_solve_docstring(PyomoSubmodelSolver_MIBS)


@Solver.register(
        name='pao.pyomo.REG',
        doc="PAO solver for Pyomo models that define linear and bilinear bilevel problems.  Solver uses regularization discussed by Scheel and Scholtes (2000) and Ralph and Wright (2004).")
class PyomoSubmodelSolver_REG(PyomoSubmodelSolverBase_LBP):
    """
    PAO REG solver for Pyomo models: pao.pyomo.REG

    This solver converts the Pyomo model to a LinearBilevelProblem and
    calls the pao.mpr.REG solver.
    """

    config = PyomoSubmodelSolverBase_LBP.config()
    config.declare('nlp_solver', ConfigValue(
        default='ipopt',
        description="The NLP solver used by REG.  (default is ipopt)"
        ))
    #config.declare('nlp_options', ConfigValue(
    #    default=None,
    #    description="A dictionary that defines the solver options for the NLP solver.  (default is None)"))
    
    def __init__(self, **kwds):
        super().__init__('pao.pyomo.REG', 'pao.mpr.REG')

    def solve(self, model, **options):
        return super().solve(model, **options)

pao.common.SolverAPI._generate_solve_docstring(PyomoSubmodelSolver_REG)


@Solver.register(
        name='pao.pyomo.PCCG',
        doc="PAO solver for Pyomo models that define linear and bilinear bilevel problems.  Solver uses projected column constraint generation algorithm described by Yue et al. (2017)")
class PyomoSubmodelSolver_PCCG(PyomoSubmodelSolverBase_LBP):
    """
    PAO PCCG solver for Pyomo models: pao.pyomo.PCCG

    This solver converts the Pyomo model to a LinearBilevelProblem and
    calls the pao.mpr.PCCG solver.
    """

    config = PyomoSubmodelSolverBase_LBP.config()
    config.declare('mip_solver', ConfigValue(
        default='cbc',
        description="The MIP solver used by PCCG.  (default is cbc)"
        ))
    #config.declare('mip_options', ConfigValue(
    #    default=None,
    #    description="A dictionary that defines the solver options for the MIP solver.  (default is None)"))
    
    def __init__(self, **kwds):
        super().__init__('pao.pyomo.PCCG', 'pao.mpr.PCCG')

    def solve(self, model, **options):
        return super().solve(model, **options)

pao.common.SolverAPI._generate_solve_docstring(PyomoSubmodelSolver_PCCG)





# pao.pyomo.examples

from . import bard511
from . import barguel
from . import besancon27
from . import getachew_ex1
from . import getachew_ex2
from . import pineda
from . import sip_example1
from . import toyexample1
from . import toyexample2
from . import toyexample3
from . import anadalingam



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao.pyomo.plugins.transform

Definition of a base class for bilevel transformation.
"""

from pyomo.core import Transformation, Var
from ..components import SubModel
import logging

logger = logging.getLogger(__name__)

class BaseBilevelTransformation(Transformation):
    """
    Base class defining methods commonly used to transform
    bilevel programs.
    """
    _fixed_vardata = dict()
    _fixed_ids = set()
    _submodel = dict()

    @property
    def fixed_vardata(self):
        return self._fixed_vardata

    @fixed_vardata.setter
    def fixed_vardata(self,key,val):
        self._fixed_vardata[key] = val

    @property
    def submodel(self):
        return self._submodel

    @submodel.setter
    def submodel(self,key,val):
        self._submodel[key] = val

    def _preprocess(self, tname, instance):
        """
        Iterate over the model collecting variable data,
        until all submodels are found.

        """
        var = {}
        instance._transformation_data[tname].submodel = list()
        for data in instance.component_objects(active=True, descend_into=True):
            name = data.name
            if isinstance(data, Var):
                var[name] = data
            elif isinstance(data, SubModel):
                submodel = data
                if submodel is None:
                    e = "Missing submodel: "+str(name)
                    logger.error(e)
                    raise RuntimeError(e)
                instance._transformation_data[tname].submodel.append(name)
                if submodel._fixed:
                    self._fixed_vardata[name] = list()
                    # if v is an indexed variable component, then append each element separately
                    for v in submodel._fixed:
                        if v.is_indexed():
                            for vardata in v.values():
                                self._fixed_vardata[name].append(vardata)
                        else:
                            self._fixed_vardata[name].append(v)
                    instance._transformation_data[tname].fixed = [v for v in self._fixed_vardata[name]]
                    self._submodel[name] = submodel
                else:
                    e = "Must specify 'fixed' or 'unfixed' options"
                    logger.error(e)
                    raise RuntimeError(e)
        return

    def _fix_all(self):
        """
        Fix the upper variables
        """
        for key,vardata in self._fixed_vardata.items():
            for v in vardata:
                if not v.fixed:
                    self._fixed_ids.add(id(v))
                    v.fixed = True

    def _unfix_all(self):
        """
        Unfix the upper variables
        """
        for key,vardata in self._fixed_vardata.items():
            for v in vardata:
                if id(v) in self._fixed_ids:
                    v.fixed = False
                    self._fixed_ids.remove(id(v))




# Example GECCO 2013 Tutorial
#
# Evolutionary Bilevel Optimization
#   A. Sinha, P. Malo, K. Deb
#   GECCO (2013)
#
# Optimal solution: (x,y) = (6,2)
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.x = pe.Var(bounds=(2,6))
    M.y = pe.Var(bounds=(0,None))

    M.o = pe.Objective(expr=M.x + 3*M.y, sense=pe.minimize)

    M.L = SubModel(fixed=M.x)
    M.L.o = pe.Objective(expr=M.y, sense=pe.maximize)
    M.L.c1 = pe.Constraint(expr= M.x + M.y <= 8)
    M.L.c2 = pe.Constraint(expr= M.x + 4*M.y >= 8)
    M.L.c3 = pe.Constraint(expr= M.x + 2*M.y <= 13)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    #M.pprint()
    opt = Solver("pao.pyomo.FA")
    opt.solve(M)
    print(pe.value(M.x))
    print(pe.value(M.y))



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao.pyomo.plugins

This package defines plugins in pao.pyomo
"""

#pylint: disable-msg=unused-import

def load():
    """
    Import the plugins defined in pao.pyomo
    """
    import pao.pyomo.solvers.mpr_solvers




#
# Example from Getachew, Mersha and Dempe, 2005.
# Constraints in the lower-level
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxR=1)
    L = U.add_lower(nxR=1)

    U.c[U] = [-1]
    U.c[L] = [-2]

    #U.b = [12, 14]

    L.c[L] = [-1]

    L.A[U] = [[-3],
              [3],
              [-2],
              [1]]
    L.A[L] = [[1], 
              [1],
              [3],
              [1]]
    L.b = [-3, 30, 12, 14]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.REG')
    opt.solve(M, tee=True)
    M.print()



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao.pyomo

This package defines modeling components for representing and solving
Pyomo models that represent multilevel problems.
"""

from pao.pyomo.components import SubModel
#from pao.pyomo.components import *
#from pao.pyomo.util import *
from .convert import convert_pyomo2LinearMultilevelProblem
from . import examples
from pao.common.solver import Solver



#
# A solver for linear bilevel programs using regularization
# discussed by Scheel and Scholtes (2000) and Ralph and Wright (2004).
#
import time
import numpy as np
import pyutilib
import pyomo.environ as pe
import pyomo.opt
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.mpec import ComplementarityList, complements

import pao.common
from ..solver import Solver, LinearMultilevelSolverBase, LinearMultilevelResults
from ..repn import LinearMultilevelProblem
from ..convert_repn import convert_to_standard_form
from . import pyomo_util


def create_model_replacing_LL_with_kkt(repn):
    """
    TODO - Document this transformation
    """
    U = repn.U
    LL = repn.U.LL
    N = len(LL)

    #
    # Create Pyomo model
    #
    M = pe.ConcreteModel()
    M.U = pe.Block()
    M.L = pe.Block(range(N))
    M.kkt = pe.Block(range(N))

    # upper- and lower-level variables
    pyomo_util.add_variables(M.U, U)
    for i in range(N):
        L = LL[i]
        # lower-level variables
        pyomo_util.add_variables(M.L[i], L)
        # dual variables
        M.kkt[i].lam = pe.Var(range(len(L.b)))                                    # equality constraints
        M.kkt[i].nu = pe.Var(range(len(L.x)), within=pe.NonNegativeReals)         # variable bounds

    # objective
    e = pyomo_util.dot(U.c[U], U.x, num=1) + U.d
    for i in range(N):
        L = LL[i]
        e += pyomo_util.dot(U.c[L], L.x, num=1)
    M.o = pe.Objective(expr=e)

    # upper-level constraints
    pyomo_util.add_linear_constraints(M.U, U.A, U, L, U.b, U.inequalities)
    for i in range(N):
        # lower-level constraints
        L = LL[i]
        pyomo_util.add_linear_constraints(M.L[i], L.A, U, L, L.b, L.inequalities)

    for i in range(N):
        L = LL[i]
        # stationarity
        M.kkt[i].stationarity = pe.ConstraintList() 
        # L_A_L' * lam
        L_A_L_T = L.A[L].transpose().todok()
        X = pyomo_util.dot( L_A_L_T, M.kkt[i].lam )
        if L.c[L] is not None:
            for k in range(len(L.c[L])):
                M.kkt[i].stationarity.add( L.c[L][k] + X[k] - M.kkt[i].nu[k] == 0 )

    for i in range(N):
        # complementarity slackness - variables
        M.kkt[i].slackness = ComplementarityList()
        for j in M.kkt[i].nu:
            M.kkt[i].slackness.add( complements( M.L[i].xR[j] >= 0, M.kkt[i].nu[j] >= 0 ) )

    return M


@Solver.register(
        name="pao.mpr.REG",
        doc="PAO solver for Multilevel Problem Representations that define linear bilevel problems.  Solver uses regularization discussed by Scheel and Scholtes (2000) and Ralph and Wright (2004).")
class LinearMultilevelSolver_REG(LinearMultilevelSolverBase):
    """
    PAO REG solver for linear MPRs: pao.mpr.REG

    This solver replaces lower-level problems using the KKT conditions and
    calls a NLP solver to solve the reformulated problem.
    """

    config = LinearMultilevelSolverBase.config()
    config.declare('nlp_solver', ConfigValue(
        default='ipopt',
        description="The NLP solver used by REG.  (default is ipopt)"
        ))
    #config.declare('nlp_options', ConfigValue(
    #    default=None,
    #    description="A dictionary that defines the solver options for the NLP solver.  (default is None)"))
    config.declare('rho', ConfigValue(
        default=1e-7,
        domain=float,
        description="The tolerance for constraints that enforce complementarity conditions.  (default is 1e-7)"
        ))

    def __init__(self, **kwds):
        super().__init__(name='pao.mpr.REG')

    def check_model(self, mpr):
        #
        # Confirm that the LinearMultilevelProblem is well-formed
        #
        assert (type(mpr) is LinearMultilevelProblem), "Solver '%s' can only solve a LinearMultilevelProblem" % self.name
        mpr.check()
        #
        # Confirm that this is a bilevel problem
        #
        for i in range(len(mpr.U.LL)):
            assert (len(mpr.U.LL[i].LL) == 0), "Can only solve bilevel problems"
        #
        # No binary or integer upper-level variables
        #
        assert (mpr.U.x.nxZ == 0), "Cannot use solver %s with model with integer upper-level variables" % self.name
        assert (mpr.U.x.nxB == 0), "Cannot use solver %s with model with binary upper-level variables" % self.name
        #
        # No binary or integer lower-level variables
        #
        for L in mpr.U.LL:
            assert (L.x.nxZ == 0), "Cannot use solver %s with model with integer lower-level variables" % self.name
            assert (L.x.nxB == 0), "Cannot use solver %s with model with binary lower-level variables" % self.name

    def solve(self, model, **options):
        #
        # Error checks
        #
        self.check_model(model)
        #
        # Process keyword options
        #
        self._update_config(options)
        #
        # Start clock
        #
        start_time = time.time()

        self.standard_form, soln_manager = convert_to_standard_form(model, inequalities=False)

        M = self._create_pyomo_model(self.standard_form, self.config.rho)
        #
        # Solve the Pyomo model the specified solver
        #
        results = LinearMultilevelResults(solution_manager=soln_manager)
        if isinstance(self.config.nlp_solver, str):
            opt = pe.SolverFactory(self.config.nlp_solver)
        else:
            opt = self.config.nlp_solver

        #if self.config.nlp_options is not None:
        #    opt.options.update(self.config.nlp_options)
        pyomo_results = opt.solve(M, tee=self.config.tee, 
                                     load_solutions=self.config.load_solutions)
        pyomo.opt.check_optimal_termination(pyomo_results)

        self._initialize_results(results, pyomo_results, M)
        results.solver.rc = getattr(opt, '_rc', None)

        if self.config.load_solutions:
            # Load results from the Pyomo model to the LinearMultilevelProblem
            results.copy_solution(From=M, To=model)
        else:
            # Load results from the Pyomo model to the Results
            results.load_from(pyomo_results)

        #self._debug()
        #results.solver.log = getattr(opt, '_log', None)

        results.solver.wallclock_time = time.time() - start_time
        return results

    def _initialize_results(self, results, pyomo_results, M):
        #
        # SOLVER
        #
        solv = results.solver
        solv.name = self.config.nlp_solver
        solv.termination_condition = pyomo_util.pyomo2pao_termination_condition(pyomo_results.solver.termination_condition)
        if hasattr(pyomo_results.solver, 'time'):
            solv.solver_time = pyomo_results.solver.time
        if self.config.load_solutions:
            solv.best_feasible_objective = pe.value(M.o)
        #
        # PROBLEM - Maybe this should be the summary of the BLP itself?
        #
        prob = results.problem
        prob.name = M.name
        prob.number_of_objectives = pyomo_results.problem.Number_of_objectives
        prob.number_of_constraints = pyomo_results.problem.Number_of_constraints
        prob.number_of_variables = pyomo_results.problem.Number_of_variables
        #prob.number_of_nonzeros = pyomo_results.problem.Number_of_nonzeros
        prob.lower_bound = pyomo_results.problem.Lower_bound
        prob.upper_bound = pyomo_results.problem.Upper_bound
        prob.sense = 'minimize'
        return results

    def _create_pyomo_model(self, repn, rho):
        M = create_model_replacing_LL_with_kkt(repn)
        #
        # Transform the problem to a MIP
        #
        xfrm = pe.TransformationFactory('mpec.simple_nonlinear')
        xfrm.apply_to(M, mpec_bound=rho)

        return M

    def _debug(self):           # pragma: no cover
        for j in M.U.xR:
            print("U",j,pe.value(M.U.xR[j]))
        for j in M.L.xR:
            print("L",j,pe.value(M.L.xR[j]))
        for j in M.kkt.lam:
            print("lam",j,pe.value(M.kkt.lam[j]))
        for j in M.kkt.nu:
            print("nu",j,pe.value(M.kkt.nu[j]))


pao.common.SolverAPI._generate_solve_docstring(LinearMultilevelSolver_REG)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao.pyomo.components

This module defines Pyomo components used to declare multi-level problems.
"""

__all__ = ("SubModel",)

from pyomo.core import SimpleBlock, ModelComponentFactory, Component
import pyomo.core.base.component_order


@ModelComponentFactory.register("A submodel in a multi-level problem")
class SubModel(SimpleBlock):
    """
    This Pyomo model component defines a sub-model in a multi-level problem.

    Pyomo models can include nested and parallel SubModel components to express 
    complex multi-level problems.
    """

    def __init__(self, *args, **kwargs):
        """
        The constructor for SubModel components.

        Keyword Args
        ------------
        fixed
            A Pyomo variable or a list of Pyomo variables that are optimized by upper-levels in the model.

        Parameters
        ----------
        *args
            Arguments passed to the SimpleBlock base class.
        **kwargs
            Other keyword arguments passed to the SimpleBlock base class.
        """
        #
        # Collect kwargs for SubModel
        #
        _rule = kwargs.pop('rule', None)
        _fixed = kwargs.pop('fixed', None)
        #
        # Initialize the SimpleBlock
        #
        kwargs.setdefault('ctype', SubModel)
        SimpleBlock.__init__(self, *args, **kwargs)
        #
        # Initialize from kwargs
        #
        self._rule = _rule
        if isinstance(_fixed, Component):
            self._fixed = [_fixed]
        else:
            self._fixed = _fixed

#
# Register 'SubModel' with Pyomo's core components.  This
# ensure that the model display() method includes SubModel
# components.
#
pyomo.core.base.component_order.display_name[SubModel] = 'SubModel'
pyomo.core.base.component_order.display_items.append(SubModel)



# 
# MIBS generalExample
#
# https://coral.ise.lehigh.edu/wp-content/uploads/2016/02/MibS_inputFile.pdf
#
# The optimal solution is C0=6, C1=5
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxZ=1)
    L = U.add_lower(nxZ=1)

    # Variables
    U.x.lower_bounds = [0.0]
    U.x.upper_bounds = [10.0]
    L.x.lower_bounds = [0.0]
    L.x.upper_bounds = [5.0]

    # Objectives
    U.c[U] = [-1]
    U.c[L] = [-7]

    L.c[L] = [1]

    # Constraints
    U.A[U] = [[-3], [1]]
    U.A[L] = [[2], [2]]
    U.b = [12, 20]

    L.A[U] = [[2], [-2]]
    L.A[L] = [[-1], [4]]
    L.b = [7, 16]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.check()
    #M.print()
    opt = Solver('pao.mpr.MIBS')
    opt.solve(M)
    print(M.U.x.values)
    print(M.U.LL.x.values)



#
# Example from Getachew, Mersha and Dempe, 2005.
# Constraints in the lower-level
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.xR = pe.Var()

    M.L = SubModel(fixed=M.xR)
    M.L.xR = pe.Var()

    M.o = pe.Objective(expr=-M.xR - 2*M.L.xR)

    M.L.o = pe.Objective(expr=-M.L.xR)
    M.L.c1 = pe.Constraint(expr= -3*M.xR +   M.L.xR <= -3)
    M.L.c2 = pe.Constraint(expr=  3*M.xR +   M.L.xR <= 30)
    M.L.c3 = pe.Constraint(expr= -2*M.xR + 3*M.L.xR <= 12)
    M.L.c4 = pe.Constraint(expr=    M.xR +   M.L.xR <= 14)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.pprint()





from pyomo.opt import *

safe_termination_conditions = [
                               TerminationCondition.maxTimeLimit,
                               TerminationCondition.maxIterations,
                               TerminationCondition.minFunctionValue,
                               TerminationCondition.minStepLength,
                               TerminationCondition.globallyOptimal,
                               TerminationCondition.locallyOptimal,
                               TerminationCondition.feasible,
                               TerminationCondition.optimal,
                               TerminationCondition.maxEvaluations,
                               TerminationCondition.other,
                              ]


#
# A solver for linear bilevel problems that
# represent interdiction problems where the upper-
# and lower-objectives are opposite.
#
import time
import numpy as np
import pyutilib
import pyomo.environ as pe
import pyomo.opt

import pao.common
from ..solver import Solver, LinearMultilevelSolverBase, LinearMultilevelResults
from ..repn import LinearMultilevelProblem
from ..convert_repn import convert_to_standard_form
from . import pyomo_util

nan = float('nan')


@Solver.register(
        name="pao.mpr.interdiction",
        doc='PAO solver for Multilevel Problem Representations that define linear interdiction problems, where the upper- and lower-objectives are opposite.')
class LinearMultilevelSolver_interdiction(LinearMultilevelSolverBase):
    """
    PAO LD solver for linear interdiction MPRs: pao.mpr.LD

    This solver replaces lower-level problems using the dual and
    calls a MIP solver to solve the reformulated problem.
    """

    def __init__(self, **kwds):
        super().__init__(name='pao.mpr.interdiction')
        self.config.solver = 'glpk'
        self.config.mipgap = nan

    def check_model(self, mpr):
        #
        # Confirm that the LinearMultilevelProblem is well-formed
        #
        assert (type(mpr) is LinearMultilevelProblem), "Solver '%s' can only solve a LinearMultilevelProblem" % self.solver_type
        mpr.check()
        #
        # TODO: For now, we just deal with the case where there is a single lower-level.  Later, we
        # will generalize this.
        #
        assert (len(mpr.L) == 1), "Only one lower-level is handled right now"
        #
        # No binary or integer lower level variables
        #
        for i in range(len(mpr.L)):
            assert (len(mpr.L[i].xZ) == 0), "Cannot use solver %s with model with integer lower-level variables" % self.solver_type
            assert (len(mpr.L[i].xB) == 0), "Cannot use solver %s with model with binary lower-level variables" % self.solver_type
        #
        # Upper and lower objectives are the opposite of each other
        #
        for i in range(len(mpr.L)):
            assert (mpr.check_opposite_objectives(mpr.U, mpr.L[i])), "Lower level L[%d] does not have an objective that is the opposite of the upper-level" % i
        #
        # Lower level variables are not allowed in the upper-level
        # constraints.
        #
        # TODO - confirm that this is a necessarily restriction for this
        # solver.
        #
        for i in range(len(mpr.L)):
            assert (U.A.L[i].xR is None), "The lower-level variables cannot be used in the upper-level constraints."
        #
        # Upper-level variables are not allowed in the lower-level
        # constraints.
        #
        # NOTE: this avoids the introduction of quadratic terms when
        # dualizing the lower-level.  That might be OK, but for now we
        # are assuming the simple application of a MIP solver.
        #
        for i in range(len(mpr.L)):
            assert (L[i].A.U.xR is None), "The upper-level variables cannot be used in the lower-level constraints."
            assert (L[i].A.U.xZ is None), "The upper-level variables cannot be used in the lower-level constraints."
            assert (L[i].A.U.xB is None), "The upper-level variables cannot be used in the lower-level constraints."

    def solve(self, mpr, options=None, **config_options):
        #
        # Error checks
        #
        self.check_model(mpr)
        #
        # Process keyword options
        #
        self._update_config(config_options)
        #
        # Start clock
        #
        start_time = time.time()

        self.standard_form, soln_manager = convert_to_standard_form(mpr)

        M = self._create_pyomo_model(self.standard_form)
        #
        # Solve the Pyomo model the specified solver
        #
        with pyomo.opt.SolverFactory(self.config.solver) as opt:
            if self.config.mipgap is not nan:
                opt.options['mipgap'] = self.config.mipgap
            if options is not None:
                opt.options.update(options)
            pyomo_results = opt.solve(M, tee=tee, 
                                         load_solutions=self.config.load_solutions)
            pyomo.opt.check_termination(pyomo_results)

            self._initialize_results(results, pyomo_results, M)
            results.solver.rc = getattr(opt, '_rc', None)

            if self.config.load_solutions:
                # Load results from the Pyomo model to the LinearMultilevelProblem
                results.copy_solution(From=M, To=mpr)
            else:
                # Load results from the Pyomo model to the Results
                results.load_from(pyomo_results)

            #self._debug()
            #results.solver.log = getattr(opt, '_log', None)

        results.solver.wallclock_time = time.time() - start_time
        return results

    def _initialize_results(self, results, pyomo_results, M):
        #
        # SOLVER
        #
        solv = results.solver
        solv.name = self.config.solver
        solv.termination_condition = pyomo_results.solver.termination_condition
        solv.solver_time = pyomo_results.solver.time
        if self.config.load_solutions:
            solv.best_feasible_objective = pe.value(M.o)
        #
        # PROBLEM - Maybe this should be the summary of the BLP itself?
        #
        prob = results.problem
        prob.name = M.name
        prob.number_of_objectives = pyomo_results.problem.Number_of_objectives
        prob.number_of_constraints = pyomo_results.problem.Number_of_constraints
        prob.number_of_variables = pyomo_results.problem.Number_of_variables
        prob.number_of_nonzeros = pyomo_results.problem.Number_of_nonzeros
        prob.lower_bound = pyomo_results.problem.Lower_bound
        prob.upper_bound = pyomo_results.problem.Upper_bound
        prob.sense = 'minimize'
        return results

    def _create_pyomo_model(self, repn):
        #
        # Create Pyomo model
        #
        M = pe.ConcreteModel()
        M.U = pe.Block()
        M.L = pe.Block()
        M.dual = pe.Block()

        # upper- and lower-level variables
        pyomo_util._create_variables(repn.U, M.U)
        pyomo_util._create_variables(repn.L, M.L)
        M.dual.z = pe.Var()

        # upper- and lower-level constraints
        pyomo_util.add_linear_constraints(M.U, repn.U.A, repn.U, repn.L, repn.U.b, M.U.inequalities)
        pyomo_util.add_linear_constraints(M.L, repn.L.A, repn.U, repn.L, repn.L.b, M.L.inequalities)

        # objective
        e = pyomo_util.dot(repn.U.c.U, repn.U, num=1) + pyomo_util.dot(repn.U.c.L, repn.L, num=1) + repn.U.d
        M.o = pe.Objective(expr=e)

        # dual variables for primal constraints
        M.Dual.dual_c = Var(range(len(M.L.c)))

        # duality gap
        e = pyomo_util._linear_expression(1, repn.c.L, repn.L)
        M.lower_gap = Constraint(expr=e[0] == M.z)

        return M

    def _debug(self):
        for j in M.U.xR:
            print("U",j,pe.value(M.U.xR[j]))
        for j in M.L.xR:
            print("L",j,pe.value(M.L.xR[j]))
        #for j in M.kkt.lam:
        #    print("lam",j,pe.value(M.kkt.lam[j]))
        #for j in M.kkt.nu:
        #    print("nu",j,pe.value(M.kkt.nu[j]))




import pyomo.environ as pe
import pao.common

Solver = pao.common.Solver


class LinearMultilevelSolverBase(pao.common.SolverAPI):
    """
    Define the API for solvers that optimize a LinearMultilevelProblem
    """

    def __init__(self, name):
        super().__init__()
        self.name = name

    def check_model(self, lmp):         # pragma: no cover
        #
        # Confirm that the LinearMultilevelProblem is well-formed
        #
        lmp.check()

    def solve(self, *args, **kwds):     # pragma: no cover
        #
        # Solve the LinearMultilevelProblem
        #
        pass


class LinearMultilevelResults(pao.common.Results):

    def __init__(self, solution_manager=None):
        super(pao.common.Results, self).__init__()
        self._solution_manager=solution_manager

    def copy_solution(self, **kwds):
        self._solution_manager.copy(**kwds)

    def load_from(self, data):          # pragma: no cover
        assert (False), "load_from() is not implemented"
        self._solution_manager.load_from(data)




#
# Toy Example 1 from 
#
# "A projection-based reformulation and decomposition algorithm for global optimization 
#  of a class of mixed integer bilevel linear programs" by Dajun Yue, Jiyao Gao, Bo Zeng, Fengqi You
#
from pao.mpr import *


def create():
    M = LinearMultilevelProblem()

    U = M.add_upper(nxZ=1)
    U.x.lower_bounds = [0]

    L = U.add_lower(nxZ=1)
    L.x.lower_bounds = [0]

    U.minimize = True
    U.c[U] = [-1]
    U.c[L] = [-10]

    L.maximize = True
    L.c[L] = [-1]

    L.A[U] = [[-25], [1], [2], [-2]]
    L.A[L] = [[20], [2], [-1], [-10]]
    L.b = [30,10,15,-15]

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    opt = Solver('pao.mpr.PCCG')
    opt.solve(M)
    M.print()



import numpy as np
import pyutilib.th as unittest
import pyomo.environ as pe
from pao.pyomo.convert import collect_multilevel_tree, convert_pyomo2LinearMultilevelProblem, convert_pyomo2MultilevelProblem
from pao.pyomo import SubModel
from pao.mpr import LinearMultilevelProblem, QuadraticMultilevelProblem


class TestMultilevelTree(unittest.TestCase):

    def test1(self):
        M = pe.ConcreteModel()
        M.x = pe.Var()
        M.o = pe.Objective(expr=M.x)
        
        var = {}
        tree = collect_multilevel_tree(M, var)

        self.assertEqual(len(tree.children), 0)
        self.assertEqual(len(var), 1)
        self.assertEqual(len(tree.orepn), 1)
        self.assertEqual(len(tree.crepn), 0)
        self.assertEqual(len(tree.fixedvars), 0)
        self.assertEqual(len(tree.unfixedvars), 1)
        self.assertEqual(len(tree.xR), 1)
        self.assertEqual(len(tree.xZ), 0)
        self.assertEqual(len(tree.xB), 0)

    def test2(self):
        # Only M.x is used in an expression
        M = pe.ConcreteModel()
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.o = pe.Objective(expr=M.x)
        
        var = {}
        tree = collect_multilevel_tree(M, var)

        self.assertEqual(len(tree.children), 0)
        self.assertEqual(len(var), 1)
        self.assertEqual(len(tree.orepn), 1)
        self.assertEqual(len(tree.crepn), 0)
        self.assertEqual(len(tree.fixedvars), 0)
        self.assertEqual(len(tree.unfixedvars), 1)
        self.assertEqual(len(tree.xR), 1)
        self.assertEqual(len(tree.xZ), 0)
        self.assertEqual(len(tree.xB), 0)

    def test3(self):
        # All variables used in an expression
        M = pe.ConcreteModel()
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.o = pe.Objective(expr=M.x+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z))
        
        var = {}
        tree = collect_multilevel_tree(M, var)

        self.assertEqual(len(tree.children), 0)
        self.assertEqual(len(var), 6)
        self.assertEqual(len(tree.orepn), 1)
        self.assertEqual(len(tree.crepn), 0)
        self.assertEqual(len(tree.fixedvars), 0)
        self.assertEqual(len(tree.unfixedvars), 6)
        self.assertEqual(len(tree.xR), 1)
        self.assertEqual(len(tree.xZ), 3)
        self.assertEqual(len(tree.xB), 2)

    def test4(self):
        # All variables used in an expression
        # One is fixed 
        M = pe.ConcreteModel()
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.z[4].fix(1)
        M.o = pe.Objective(expr=M.x+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z))
        
        var = {}
        tree = collect_multilevel_tree(M, var)

        self.assertEqual(len(tree.children), 0)
        self.assertEqual(len(var), 5)
        self.assertEqual(len(tree.orepn), 1)
        self.assertEqual(len(tree.crepn), 0)
        self.assertEqual(len(tree.fixedvars), 0)
        self.assertEqual(len(tree.unfixedvars), 5)
        self.assertEqual(len(tree.xR), 1)
        self.assertEqual(len(tree.xZ), 2)
        self.assertEqual(len(tree.xB), 2)

    def test5(self):
        # All variables used in an expression
        # One of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.z[4].setlb(0)
        M.z[4].setub(1)
        M.o = pe.Objective(expr=M.x+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z))
        
        var = {}
        tree = collect_multilevel_tree(M, var)

        self.assertEqual(len(tree.children), 0)
        self.assertEqual(len(var), 6)
        self.assertEqual(len(tree.orepn), 1)
        self.assertEqual(len(tree.crepn), 0)
        self.assertEqual(len(tree.fixedvars), 0)
        self.assertEqual(len(tree.unfixedvars), 6)
        self.assertEqual(len(tree.xR), 1)
        self.assertEqual(len(tree.xZ), 2)
        self.assertEqual(len(tree.xB), 3)

    def test5a(self):
        # All variables used in an expression
        # One of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.lb = pe.Param(initialize=0)
        M.ub = pe.Param(initialize=1)
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.z[4].setlb(M.lb)
        M.z[4].setub(M.ub)
        M.o = pe.Objective(expr=M.x+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z))
        
        var = {}
        tree = collect_multilevel_tree(M, var)

        self.assertEqual(len(tree.children), 0)
        self.assertEqual(len(var), 6)
        self.assertEqual(len(tree.orepn), 1)
        self.assertEqual(len(tree.crepn), 0)
        self.assertEqual(len(tree.fixedvars), 0)
        self.assertEqual(len(tree.unfixedvars), 6)
        self.assertEqual(len(tree.xR), 1)
        self.assertEqual(len(tree.xZ), 2)
        self.assertEqual(len(tree.xB), 3)

    def test6a(self):
        # All variables used in an expression
        # One of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.p = pe.Param(initialize=1)
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.z[4].setlb(0)
        M.z[4].setub(1)
        M.o = pe.Objective(expr=M.x+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z))
        M.c1 = pe.Constraint(expr=M.p <= 0)
        
        var = {}
        try:
            tree = collect_multilevel_tree(M, var)
            self.fail("Expected AssertionError")
        except AssertionError:
            pass

    def test6b(self):
        # All variables used in an expression
        # One of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.p = pe.Param(initialize=1)
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.z[4].setlb(0)
        M.z[4].setub(1)
        M.o = pe.Objective(expr=M.x+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z))
        M.c1 = pe.Constraint(expr=pe.inequality(2, M.p, 3))
        
        var = {}
        try:
            tree = collect_multilevel_tree(M, var)
            self.fail("Expected AssertionError")
        except AssertionError:
            pass

    def test6c(self):
        # All variables used in an expression
        # One of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.p = pe.Param(initialize=1)
        M.x = pe.Var()
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.z[4].setlb(0)
        M.z[4].setub(1)
        M.o = pe.Objective(expr=M.x+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z))
        M.c1 = pe.Constraint(expr=M.p == 0)
        
        var = {}
        try:
            tree = collect_multilevel_tree(M, var)
            self.fail("Expected AssertionError")
        except AssertionError:
            pass

    def test_initialize1(self):
        # All variables used in an expression
        # One of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.x = pe.Var()
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z[4].setlb(0)
        M.z[4].setub(1)
        M.o = pe.Objective(expr=1+2*M.x+3*sum(M.y[i] for i in M.y)+4*sum(M.z[i] for i in M.z))
        
        M.s = SubModel(fixed=M.x)
        M.s.o = pe.Objective(expr=5+6*M.x+7*sum(M.y[i] for i in M.y)+8*sum(M.z[i] for i in M.z))

        lmpr,_ = convert_pyomo2LinearMultilevelProblem(M, inequalities=True)
        
        U = lmpr.U
        L = lmpr.U.LL

        self.assertEqual(len(U.x), len(M.x))
        self.assertEqual(len(L.x), len(M.z)+len(M.y))

        self.assertEqual(U.d, 1)
        self.assertEqual(list(U.c[U]), [2])
        self.assertEqual(list(U.c[L]), [4,4,3,3,4])

        self.assertEqual(L.d, 5)
        self.assertEqual(list(L.c[U]), [6])
        self.assertEqual(list(L.c[L]), [8,8,7,7,8])

    def test_initialize_2(self):
        # All variables used in an expression
        # None of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.p = pe.Param(initialize=1)
        M.x = pe.Var([0])
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.x[0].setlb(0)
        M.x[0].setub(1)
        M.z[4].setlb(0)
        M.z[4].setub(2)
        M.o = pe.Objective(expr=M.x[0]+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z))
        def c_rule(M, i):
            return pe.Constraint.Skip
        M.c1 = pe.Constraint([0], rule=c_rule)
        M.c2 = pe.Constraint(expr=M.x[0] <= np.PINF)
        M.c3 = pe.Constraint(expr=M.p <= 2)
        
        M.s = SubModel(fixed=M.x)

        lmpr,_ = convert_pyomo2LinearMultilevelProblem(M, inequalities=True)
        
        U = lmpr.U
        L = lmpr.U.LL

        self.assertEqual(len(U.x), len(M.x)+len(M.z)+len(M.y))
        self.assertEqual(U.d, 0)

        self.assertEqual(list(U.c[U]), [1,3,3,3,2,2])

    def test_initialize_2L(self):
        # All variables used in an expression
        # None of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.p = pe.Param(initialize=1)
        M.x = pe.Var([0])
        M.X = pe.Var([0])
        M.y = pe.Var([1,2], within=pe.Binary)
        M.Y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.Z = pe.Var([3,4,5], within=pe.Integers)
        M.x[0].setlb(0)
        M.x[0].setub(1)
        M.z[4].setlb(0)
        M.z[4].setub(2)
        M.o = pe.Objective(expr=M.x[0]+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z))
        def c_rule(M, i):
            return pe.Constraint.Skip
        M.c1 = pe.Constraint([0], rule=c_rule)
        M.c2 = pe.Constraint(expr=M.x[0] <= np.PINF)
        M.c3 = pe.Constraint(expr=M.p <= 2)
        
        M.s = SubModel(fixed=M.x)

        M.s.o = pe.Objective(expr=M.X[0]+2*sum(M.Y[i] for i in M.Y)+3*sum(M.Z[i] for i in M.Z))

        M.s.c1 = pe.Constraint([0], rule=c_rule)
        M.s.c2 = pe.Constraint(expr=M.x[0] <= np.PINF)
        M.s.c3 = pe.Constraint(expr=M.p <= 2)

        lmpr,_ = convert_pyomo2LinearMultilevelProblem(M, inequalities=True)
        
        U = lmpr.U
        L = lmpr.U.LL

        self.assertEqual(len(U.x), len(M.x)+len(M.z)+len(M.y))
        self.assertEqual(len(L.x), len(M.X)+len(M.Z)+len(M.Y))

        self.assertEqual(U.d, 0)

        self.assertEqual(list(U.c[U]), [1,3,3,3,2,2])
        self.assertEqual(U.c[L], None)

        self.assertEqual(L.c[U], None)
        self.assertEqual(list(L.c[L]), [1,3,3,3,2,2])

    def test_initialize_3(self):
        # All variables used in an expression
        # None of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.x = pe.Var([0])
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.x[0].setlb(0)
        M.x[0].setub(1)
        M.z[4].setlb(0)
        M.z[4].setub(2)
        M.o = pe.Objective(expr=1)
        M.c = pe.Constraint(expr=M.x[0]+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z) == 0)
        
        M.s = SubModel(fixed=M.x)
        #M.s.c = pe.Constraint(expr=M.x[0]+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z) == 0)

        lmpr,_ = convert_pyomo2LinearMultilevelProblem(M, inequalities=True)
        
        U = lmpr.U
        L = lmpr.U.LL

        self.assertEqual(len(U.x), len(M.x)+len(M.z)+len(M.y))

        self.assertEqual(U.c[U], None)

        # 2 rows because the lmpr is an inequality, so the equality is split in two
        self.assertEqual(U.A[U].shape, (2,6))
        self.assertEqual([list(U.A[U].toarray()[i]) for i in range(2)], [[1,3,3,3,2,2], [-1,-3,-3,-3,-2,-2]])

    def test_initialize_3a(self):
        # All variables used in an expression
        # None of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.x = pe.Var([0])
        M.y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.x[0].setlb(0)
        M.x[0].setub(1)
        M.z[4].setlb(0)
        M.z[4].setub(2)
        M.o = pe.Objective(expr=1)
        M.c = pe.Constraint(expr=M.x[0]+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z) == 0)
        
        M.s = SubModel(fixed=M.x)
        #M.s.c = pe.Constraint(expr=M.x[0]+sum(M.y[i] for i in M.y)+sum(M.z[i] for i in M.z) == 0)

        lmpr,_ = convert_pyomo2LinearMultilevelProblem(M, inequalities=False)
        
        U = lmpr.U
        L = lmpr.U.LL

        self.assertEqual(len(lmpr.U.x), len(M.x)+len(M.z)+len(M.y))

        self.assertEqual(U.c[U], None)

        # 2 rows because the lmpr is an inequality, so the equality is split in two
        self.assertEqual(U.A[U].shape, (1,6))
        self.assertEqual([list(U.A[U].toarray()[i]) for i in range(1)], [[1,3,3,3,2,2]])

    def test_initialize_4a(self):
        # All variables used in an expression
        # None of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.x = pe.Var([0])
        M.X = pe.Var([0])
        M.y = pe.Var([1,2], within=pe.Binary)
        M.Y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.Z = pe.Var([3,4,5], within=pe.Integers)
        M.x[0].setlb(0)
        M.x[0].setub(1)
        M.z[4].setlb(0)
        M.z[4].setub(2)
        M.o = pe.Objective(expr=1)
        M.c = pe.Constraint(expr=M.x[0]+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z) == 0)
        M.C1 = pe.Constraint(expr=M.x[0] >= 0)
        M.C2 = pe.Constraint(expr=M.x[0] <= 0)
        M.C3 = pe.Constraint(expr=pe.inequality(0, M.x[0], 1))
        
        M.s = SubModel(fixed=M.x)
        M.s.c = pe.Constraint(expr=M.X[0]+sum(M.Y[i] for i in M.Y)+sum(M.Z[i] for i in M.Z) == 0)

        lmpr,_ = convert_pyomo2LinearMultilevelProblem(M, inequalities=True)
        
        U = lmpr.U
        L = U.LL

        self.assertEqual(len(U.x), len(M.x)+len(M.z)+len(M.y))
        self.assertEqual(len(L.x), len(M.X)+len(M.Z)+len(M.Y))

        self.assertEqual(U.c[U], None)

        # 2 rows because the lmpr is an inequality, so the equality is split in two
        self.assertEqual(U.A[U].shape, (6,6))
        self.assertEqual([list(U.A[U].toarray()[i]) for i in range(6)],
[[1.0,3,3,3,2,2], [-1.0,-3,-3,-3,-2,-2], [-1.0,0,0,0,0,0], [1.0,0,0,0,0,0], [-1.0,0,0,0,0,0], [1.0,0,0,0,0,0]])

        # 2 rows because the lmpr is an inequality, so the equality is split in two
        self.assertEqual(L.A[L].shape, (2,6))
        self.assertEqual([list(L.A[L].toarray()[i]) for i in range(2)], [[1,1,1,1,1,1],[-1,-1,-1,-1,-1,-1]])

    def test_initialize_4b(self):
        # All variables used in an expression
        # None of the Integer variables looks like a binary
        M = pe.ConcreteModel()
        M.x = pe.Var([0])
        M.X = pe.Var([0])
        M.y = pe.Var([1,2], within=pe.Binary)
        M.Y = pe.Var([1,2], within=pe.Binary)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.Z = pe.Var([3,4,5], within=pe.Integers)
        M.x[0].setlb(0)
        M.x[0].setub(1)
        M.z[4].setlb(0)
        M.z[4].setub(2)
        M.o = pe.Objective(expr=1)
        M.c = pe.Constraint(expr=M.x[0]+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z) == 0)
        M.C1 = pe.Constraint(expr=M.x[0] >= 0)
        M.C2 = pe.Constraint(expr=M.x[0] <= 0)
        M.C3 = pe.Constraint(expr=pe.inequality(0, M.x[0], 1))
        
        M.s = SubModel(fixed=M.x)
        M.s.c = pe.Constraint(expr=M.X[0]+sum(M.Y[i] for i in M.Y)+sum(M.Z[i] for i in M.Z) == 0)

        lmpr,_ = convert_pyomo2LinearMultilevelProblem(M, inequalities=False)
        
        U = lmpr.U
        L = U.LL

        self.assertEqual(len(U.x), len(M.x)+4+len(M.z)+len(M.y))
        self.assertEqual(len(L.x), len(M.X)+len(M.Z)+len(M.Y))

        self.assertEqual(U.c[U], None)

        # 2 rows because the lmpr is an inequality, so the equality is split in two
        self.assertEqual(U.A[U].shape, (5,10))
        self.assertEqual([list(U.A[U].toarray()[i]) for i in range(5)],
[[ 1.0, 0.0, 0.0, 0.0, 0.0, 3, 3, 3, 2, 2],
 [-1.0, 1.0, 0.0, 0.0, 0.0, 0, 0, 0, 0, 0],
 [ 1.0, 0.0, 1.0, 0.0, 0.0, 0, 0, 0, 0, 0],
 [-1.0, 0.0, 0.0, 1.0, 0.0, 0, 0, 0, 0, 0],
 [ 1.0, 0.0, 0.0, 0.0, 1.0, 0, 0, 0, 0, 0]])

        # 2 rows because the lmpr is an inequality, so the equality is split in two
        self.assertEqual(L.A[L].shape, (1,6))
        self.assertEqual([list(L.A[L].toarray()[i]) for i in range(1)], [[1,1,1,1,1,1]])

    def test_initialize_6(self):
        # Bilinear terms in objective
        M = pe.ConcreteModel()
        M.x = pe.Var(within=pe.Binary)
        M.y = pe.Var([1,2], within=pe.Reals)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.o = pe.Objective(expr=2*M.z[3] + 3*M.x*M.y[1])
        M.c = pe.Constraint(expr=M.x+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z) <= 0)
        
        M.s = SubModel(fixed=M.x)
        M.s.o = pe.Objective(expr=4*M.z[3] + 3*M.y[1] - 5*M.x*M.y[2])

        qmp,_ = convert_pyomo2MultilevelProblem(M)
        #qmp.print()
        
        self.assertEqual(type(qmp), QuadraticMultilevelProblem)
        U = qmp.U
        L = qmp.U.LL

        self.assertEqual(len(U.x), 3)
        self.assertEqual(len(L.x), 3)

        self.assertEqual(U.c[U], None)
        self.assertEqual(list(U.c[L]), [0,0,2])
        self.assertEqual(U.P[U,L].toarray().tolist(),
[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [3.0, 0.0, 0.0]])

        self.assertEqual(L.c[U], None)
        self.assertEqual(list(L.c[L]), [3,0,4])
        self.assertEqual(L.P[U,L].toarray().tolist(),
[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, -5.0, 0.0]])

        self.assertEqual(U.A[U].toarray().tolist(),
[[3.0, 3.0, 1.0]])
        self.assertEqual(U.A[L].toarray().tolist(), [[2,2,3]])

        self.assertEqual(L.A[U], None)
        self.assertEqual(L.A[L], None)

    def test_initialize_7(self):
        # Bilinear terms in objective
        M = pe.ConcreteModel()
        M.x = pe.Var(within=pe.Binary)
        M.y = pe.Var([1,2], within=pe.Reals)
        M.z = pe.Var([3,4,5], within=pe.Integers)
        M.o = pe.Objective(expr=2*M.z[3] + 3*M.x*M.y[1])
        M.c = pe.Constraint(expr=M.x+2*sum(M.y[i] for i in M.y)+3*sum(M.z[i] for i in M.z) + 7*M.x*M.y[1] <= 0)
        
        M.s = SubModel(fixed=M.x)
        M.s.o = pe.Objective(expr=4*M.z[3] + 3*M.y[1] - 5*M.x*M.y[2])
        M.s.c = pe.Constraint(expr=11*M.x*M.z[3] == 0)

        qmp,_ = convert_pyomo2MultilevelProblem(M)
        #qmp.print()
        
        self.assertEqual(type(qmp), QuadraticMultilevelProblem)
        U = qmp.U
        L = qmp.U.LL

        self.assertEqual(len(U.x), 3)
        self.assertEqual(len(L.x), 3)

        self.assertEqual(U.c[U], None)
        self.assertEqual(list(U.c[L]), [0,0,2])
        self.assertEqual(U.P[U,L].toarray().tolist(),
[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [3.0, 0.0, 0.0]])

        self.assertEqual(L.c[U], None)
        self.assertEqual(list(L.c[L]), [3,0,4])
        self.assertEqual(L.P[U,L].toarray().tolist(),
[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, -5.0, 0.0]])

        self.assertEqual(U.A[U].toarray().tolist(),
[[3.0, 3.0, 1.0]])
        self.assertEqual(U.A[L].toarray().tolist(), [[2,2,3]])
        self.assertEqual(U.Q[U,L][0].toarray().tolist(),
[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [7.0, 0.0, 0.0]])

        self.assertEqual(L.A[U], None)
        self.assertEqual(L.A[L], None)
        self.assertEqual(L.Q[U,L][0].toarray().tolist(),
[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 11.0]])
        self.assertEqual(L.Q[U,L][1].toarray().tolist(),
[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, -11.0]])

if __name__ == "__main__":
    unittest.main()



#
# Example from Pineda and Morales (2018), showing how
# bad estimates of bigMs give the wrong answer.
#
import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()

    M.xR = pe.Var(bounds=(0,2))

    M.L = SubModel(fixed=(M.xR))
    M.L.xR = pe.Var(bounds=(0,None))

    M.o = pe.Objective(expr=M.xR + M.L.xR, sense=pe.maximize)

    M.L.o = pe.Objective(expr=M.L.xR)
    M.L.c = pe.Constraint(expr=100*M.xR - M.L.xR <= 100)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.pprint()



# Example 5.1.1 from
#
# Practical Bilevel Optimization: Algorithms and Applications
#   Jonathan Bard

import pyomo.environ as pe
from pao.pyomo import *


def create():
    M = pe.ConcreteModel()
    M.x = pe.Var(bounds=(0,None))
    M.y = pe.Var(bounds=(0,None))

    M.o = pe.Objective(expr=M.x - 4*M.y)

    M.L = SubModel(fixed=M.x)
    M.L.o = pe.Objective(expr=M.y)
    M.L.c1 = pe.Constraint(expr=   -M.x -   M.y <= -3)
    M.L.c2 = pe.Constraint(expr= -2*M.x +   M.y <=  0)
    M.L.c3 = pe.Constraint(expr=  2*M.x +   M.y <= 12)
    M.L.c4 = pe.Constraint(expr=  3*M.x - 2*M.y <=  4)
    #M.L.c4 = pe.Constraint(expr= -3*M.x + 2*M.y <= -4)

    return M


if __name__ == "__main__":          #pragma: no cover
    M = create()
    M.pprint()



"""
Script to generate the installer for pao.
"""

import sys
import os


def _find_packages(path):
    """
    Generate a list of nested packages
    """
    pkg_list = []
    if not os.path.exists(path):
        return []
    if not os.path.exists(path+os.sep+"__init__.py"):
        return []
    else:
        pkg_list.append(path)
    for root, dirs, files in os.walk(path, topdown=True):
        if root in pkg_list and "__init__.py" in files:
            for name in dirs:
                if os.path.exists(root+os.sep+name+os.sep+"__init__.py"):
                    pkg_list.append(root+os.sep+name)
    return [pkg for pkg in map(lambda x:x.replace(os.sep, "."), pkg_list)]


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

requires = [
    'Pyomo',
    'parameterized',
    'scipy',
    'munch'
    ]

from setuptools import setup
import sys

packages = _find_packages('pao')

setup(name='pao',
      version='1.0.1',
      maintainer='William E. Hart',
      maintainer_email='wehart@sandia.gov',
      url='https://github.com/or-fusion/pao',
      license='BSD',
      platforms=["any"],
      description='PAO: Python Adversarial Optimization',
      long_description=read('README.txt'),
      classifiers=[
        #'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        #'Programming Language :: Python :: Implementation :: Jython',
        #'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules' ],
      packages=packages,
      keywords=['optimization'],
      install_requires=requires,
      python_requires='>=3.7',
      )



# coding=utf-8
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

"""
pao.pyomo.plugins

This module defines the collect_bilevel_matrix_representation() function, which
collects information that is used to in bilevel algorithm development.
"""

from pyutilib.misc import Bunch
from pyomo.repn import generate_standard_repn
from pyomo.core import (Var,
                        Constraint,
                        Objective,
                        minimize,
                        maximize,
                        NonNegativeReals,
                        NonPositiveReals,
                        Reals,
                        Block,
                        Model,
                        ConcreteModel)
from scipy.sparse import coo_matrix
from ..components import SubModel
from numpy import zeros, array
from collections import OrderedDict


def collect_global_vars(model):
    """
    Collect vars on the model
    """
    c_var_ids = set()
    b_var_ids = set()
    i_var_ids = set()

    all_vars = dict()
    for v in model.component_objects(Var, descend_into=False, sort=True):
        if v.is_indexed():
            for vardata in v.values():
                all_vars[id(vardata)] = vardata
                if v.is_continuous():
                    c_var_ids.add(id(vardata))
                if v.is_binary():
                    b_var_ids.add(id(vardata))
                if v.is_integer():
                    i_var_ids.add(id(vardata))
        else:
            all_vars[id(v)] = v
            if v.is_continuous():
                c_var_ids.add(id(v))
            if v.is_binary():
                b_var_ids.add(id(v))
            if v.is_integer():
                i_var_ids.add(id(v))

    return all_vars, c_var_ids, b_var_ids, i_var_ids


def collect_fixed_vars(block):
    """
    Collect fixed vars on the block
    """
    fixed_var_ids = set()
    if hasattr(block, '_fixed'):
        for v in block._fixed:
            if v.is_indexed():
                for vardata in v.values():
                    fixed_var_ids.add(id(vardata))
            else:
                fixed_var_ids.add(id(v))

    return fixed_var_ids


def collect_bilevel_objective_vector_representation(block, all_vars, c_var_ids, b_var_ids, i_var_ids, fixed_var_ids,
                                                    standard_form=True):
    """
    Collects the terms for the objective
    for the given local variable scope, per variable type (c:continuous,
    b:binary, i:integer). Nested Blocks (or SubModels) are not considered
    within local scope.

    Arguments:
        block: A Pyomo Model, Block, or SubModel that will be processed to
                collect the objective terms.
        all_vars: An iterable, sorted object with Variable and VarData values.
        c_var_ids: Unique id for vars that are continuous and not fixed on this block.
        b_var_ids: Unique id for vars that are binary and not fixed on this block.
        i_var_ids: Unique id for vars that are integer and not fixed on this block.
        fixed_var_ids: Unique id for vars that are fixed on this block.
        standard_form: Whether to put the model into stanndard form representation (max for subproblem; min for model).

    Returns: Tuple with the following values:
    C^{c}:         The C coefficient scalar associate to continuous vars
    C_q^{c}:       The C coefficient vector associate to continuous vars in bilinear terms
    C^{b}:         The C coefficient scalar associate to binary vars
    C_q^{b}:       The C coefficient vector associate to binary vars in bilinear terms
    C^{i}:         The C coefficient scalar associate to integer vars
    C_q^{i}:       The C coefficient vector associate to integer vars in bilinear terms
    C^{f}:         The C coefficient scalar associate to fixed vars
    C_q^{f}:       The C coefficient vector associate to fixed vars in bilinear terms
    C_constant:    The C coefficient vector associate to constant terms
    """

    # number of variables
    _c = len(all_vars)

    # C coefficient for linear terms, continuous var
    C_c = {k: 0. for k in c_var_ids}

    # C coefficients for bilinear terms, for the continuous var
    C_c_q = {k: zeros(_c) for k in c_var_ids}

    # C coefficient for linear terms, binary var
    C_b = {k: 0. for k in b_var_ids}

    # C coefficients for bilinear terms, for the binary var
    C_b_q = {k: zeros(_c) for k in b_var_ids}

    # C coefficient for linear terms, integer var
    C_i = {k: 0. for k in i_var_ids}

    # C coefficients for bilinear terms, for the integer var
    C_i_q = {k: zeros(_c) for k in i_var_ids}

    # C coefficient for linear terms, fixed var in submodel
    C_f = {k: 0. for k in fixed_var_ids}

    # C coefficients for bilinear terms, fixed var in submodel
    C_f_q = {k: zeros(_c) for k in fixed_var_ids}

    C_constant = 0.

    nobj = 0
    for odata in block.component_objects(Objective, active=True):
        for ndx in odata:
            o_terms = generate_standard_repn(odata[ndx].expr, compute_values=False)
            C_constant = o_terms.constant
            _sign = 1.
            if odata[ndx].sense == minimize: # swith to a maximize lower-level
                if standard_form and not block.parent_block() is None:
                    _sign = -1.
            if odata[ndx].sense == maximize: # switch to a minimize upper-level
                if standard_form and id(block) == id(block.root_block()):
                    _sign = -1.
            for var, coef in zip(o_terms.linear_vars, o_terms.linear_coefs):
                _vid = id(var)
                if _vid in c_var_ids:
                    C_c[_vid] = _sign * coef
                if _vid in b_var_ids:
                    C_b[_vid] = _sign * coef
                if _vid in i_var_ids:
                    C_i[_vid] = _sign * coef
                if _vid in fixed_var_ids:
                    C_f[_vid] = _sign * coef

            for (var1, var2), coef in zip(o_terms.quadratic_vars, o_terms.quadratic_coefs):
                for (var, fixed) in [(var1, var2), (var2, var1)]:
                    _vid = id(var)
                    _col = list(all_vars.keys()).index(id(fixed)) if (id(fixed) in fixed_var_ids or id(var) in fixed_var_ids) else None
                    if not _col is None:
                        if _vid in c_var_ids:
                            C_c_q[_vid][_col] = _sign * coef
                        if _vid in b_var_ids:
                            C_b_q[_vid][_col] = _sign * coef
                        if _vid in i_var_ids:
                            C_i_q[_vid][_col] = _sign * coef
                        if _vid in fixed_var_ids:
                            C_f_q[_vid][_col] = _sign * coef
                    else:
                        raise RuntimeError(
                            "Error in matrix representation of bilevel Submodel block.  Bilinear expressions using"
                            "variables in the same bilevel upper- or lower-level.")
            nobj += 1
    if nobj == 0:
        raise RuntimeError("Error in vector representation of bilevel Submodel block.  No objective expression.")
    if nobj > 1:
        raise RuntimeError("Error in vector representation of bilevel Submodel block.  Multiple objective expressions.")

    return C_c, C_c_q, C_b, C_b_q, C_i, C_i_q, C_f, C_f_q, C_constant

def collect_bilevel_constraint_matrix_representation(block, all_vars, c_var_ids, b_var_ids, i_var_ids, fixed_var_ids,
                                                     standard_form=True):
    """
    Collects the terms for each ==, >= and <= constraints
    for the given local variable scope, per variable type (c:continuous,
    b:binary, i:integer). Nested Blocks (or SubModels) are not considered
    within local scope.

    TODO: The way the coo_matrix is constructed (by casting a numpy matrix)
    TODO: may be resource intensive. If that is the case, we will want to update the code
    TODO: as follows:
    # Constructing a coo_matrix using ijv format
    row  = np.array([r1, ...., rN])
    col  = np.array([c1, ...., cN])
    data = np.array([d1, ...., dN])
    _r = len(cons_sense_rhs)
    _c = len(all_vars)
    coo_matrix((data, (row, col)), shape=(_r, _c))

    Arguments:
        block: A Pyomo Model, Block, or SubModel that will be processed to
                collect terms for Ax <= b, Ax >= b, and Ax == b.
        all_vars: An iterable, sorted object with Variable and VarData values.
        c_var_ids: Unique id for vars that are continuous and not fixed on this block.
        b_var_ids: Unique id for vars that are binary and not fixed on this block.
        i_var_ids: Unique id for vars that are integer and not fixed on this block.
        fixed_var_ids: Unique id for vars that are fixed on this block.
        standard_form: Whether to put the model into stanndard form representation (<=, ==).

    Returns: Tuple with the following values:
    A^{c}:         The A coefficient vector associate to continuous vars
    A_q^{c}:       The A coefficient matrix associate to continuous vars in bilinear terms
    A^{b}:         The A coefficient vector associate to binary vars
    A_q^{b}:       The A coefficient matrix associate to binary vars in bilinear terms
    A^{i}:         The A coefficient vector associate to integer vars
    A_q^{i}:       The A coefficient matrix associate to integer vars in bilinear terms
    A^{f}:         The A coefficient vector associate to fixed vars
    A_q^{f}:       The A coefficient matrix associate to fixed vars in bilinear terms
    cons_sense_rhs: Unique id for constraints on the block.
    """

    # number of constraints
    _r = 0
    for data in block.component_map(Constraint, active=True).itervalues():
        _r += len(data)
        if data.has_lb() and data.has_ub():
            _r += len(data)

    # number of variables
    _c = len(all_vars)

    # A coefficients for linear terms, continuous var
    A_c = {k: zeros(_r) for k in c_var_ids}

    # A coefficients for bilinear terms, for the continuous var
    A_c_q = {k: zeros((_r, _c)) for k in c_var_ids}

    # A coefficients for linear terms, binary var
    A_b = {k: zeros(_r) for k in b_var_ids}

    # A coefficients for bilinear terms, for the binary var
    A_b_q = {k: zeros((_r, _c)) for k in b_var_ids}

    # A coefficients for linear terms, integer var
    A_i = {k: zeros(_r) for k in i_var_ids}

    # A coefficients for bilinear terms, for the integer var
    A_i_q = {k: zeros((_r, _c)) for k in i_var_ids}

    # A coefficients for linear terms, fixed var in submodel
    A_f = {k: zeros(_r) for k in fixed_var_ids}

    # A coefficients for bilinear terms, fixed var in submodel
    A_f_q = {k: zeros((_r, _c)) for k in fixed_var_ids}

    cons_sense_rhs = OrderedDict()
    for c in block.component_objects(Constraint, active=True, descend_into=True):
        sense = list()
        for ndx in c:
            con = c[ndx]
            _cid = id(con)

            lower_terms = generate_standard_repn(con.lower, compute_values=False) \
                if not con.lower is None else None
            upper_terms = generate_standard_repn(con.upper, compute_values=False) \
                if not con.upper is None else None
            body_terms = generate_standard_repn(con.body, compute_values=True)
            if con.equality:
                sense.append('e')
                cons_sense_rhs[id(con), 'e'] = lower_terms.constant - body_terms.constant
            else:
                if not (lower_terms is None):
                    if standard_form:
                        sense.append('g->l')
                        cons_sense_rhs[id(con), 'g->l'] = -(lower_terms.constant - body_terms.constant)
                    else:
                        sense.append('g')
                        cons_sense_rhs[id(con), 'g'] = lower_terms.constant - body_terms.constant
                if not (upper_terms is None):
                    sense.append('l')
                    cons_sense_rhs[id(con), 'l'] = upper_terms.constant - body_terms.constant
            for s in sense:
                _sign = 1.
                if s=='g->l':
                    _sign = -1.
                _row = list(cons_sense_rhs.keys()).index((_cid, s))

                for var, coef in zip(body_terms.linear_vars, body_terms.linear_coefs):
                    _vid = id(var)
                    if _vid in c_var_ids:
                        A_c[_vid][_row] = _sign*coef
                    if _vid in b_var_ids:
                        A_b[_vid][_row] = _sign*coef
                    if _vid in i_var_ids:
                        A_i[_vid][_row] = _sign*coef
                    if _vid in fixed_var_ids:
                        A_f[_vid][_row] = _sign*coef

                for (var1, var2), coef in zip(body_terms.quadratic_vars, body_terms.quadratic_coefs):
                    for (var, fixed) in [(var1, var2), (var2, var1)]:
                        _vid = id(var)
                        _col = list(all_vars.keys()).index(id(fixed)) if (id(fixed) in fixed_var_ids or id(var) in fixed_var_ids) else None
                        if not _col is None:
                            if _vid in c_var_ids:
                                A_c_q[_vid][_row, _col] = _sign*coef
                            if _vid in b_var_ids:
                                A_b_q[_vid][_row, _col] = _sign*coef
                            if _vid in i_var_ids:
                                A_i_q[_vid][_row, _col] = _sign*coef
                            if _vid in fixed_var_ids:
                                A_f_q[_vid][_row, _col] = _sign*coef
                        else:
                            raise RuntimeError(
                                "Error in matrix representation of bilevel Submodel block.  Bilinear expressions using"
                                " variables in the same bilevel upper- or lower-level.")

    A_c_q = {k: coo_matrix(v) for k, v in A_c_q.items()}
    A_b_q = {k: coo_matrix(v) for k, v in A_b_q.items()}
    A_i_q = {k: coo_matrix(v) for k, v in A_i_q.items()}
    A_f_q = {k: coo_matrix(v) for k, v in A_f_q.items()}

    return A_c, A_c_q, A_b, A_b_q, A_i, A_i_q, A_f, A_f_q, cons_sense_rhs


class BilevelMatrixRepn():
    """
    A class that creates an object which contains all the matrix expressions
    to the Pyomo model.
    """

    def __init__(self, model, **kwds):
        # Put in Standard Form
        self._standard_form = kwds.pop('standard_form', True)

        # Pyomo ConcreteModel
        self._model = model

        # List of model block name and all Submodels
        self._all_models = {block.name for block in model.component_objects(SubModel)}
        self._all_models.add(model.name)

        # All variables on model, including var id sets for continuous (c), binary (b) and integer (i) variable types
        all_vars, c_var_ids, b_var_ids, i_var_ids = collect_global_vars(model)
        self._all_vars = all_vars
        self._c_var_ids = c_var_ids
        self._b_var_ids = b_var_ids
        self._i_var_ids = i_var_ids

        # All fixed variables that exist on specified SubModel blocks (these do not exist on the Model block)
        self._fixed_var_ids = {block.name: collect_fixed_vars(block) for block in model.component_objects(SubModel)}

        # A coefficients for linear terms, continuous var
        self._A_c = {k: dict() for k in self._all_models}
        # A coefficients for bilinear terms, for the continuous var
        self._A_c_q = {k: dict() for k in self._all_models}
        # A coefficients for linear terms, binary var
        self._A_b = {k: dict() for k in self._all_models}
        # A coefficients for bilinear terms, for the binary var
        self._A_b_q = {k: dict() for k in self._all_models}
        # A coefficients for linear terms, integer var
        self._A_i = {k: dict() for k in self._all_models}
        # A coefficients for bilinear terms, for the integer var
        self._A_i_q = {k: dict() for k in self._all_models}
        # A coefficients for linear terms, fixed var in SubModel
        self._A_f = {k: dict() for k in self._all_models - {model.name}}
        # A coefficients for bilinear terms, fixed var in SubModel
        self._A_f_q = {k: dict() for k in self._all_models - {model.name}}
        # All constraint ids, corresponding sense and rhs; includes two entries for constraint declarations LHS <= body <= RHS
        # that are not equality constraints
        self._cons_sense_rhs = {k: dict() for k in self._all_models}

        # C coefficients for linear terms, continuous var
        self._C_c = {k: dict() for k in self._all_models}
        # C coefficients for bilinear terms, for the continuous var
        self._C_c_q = {k: dict() for k in self._all_models}
        # C coefficients for linear terms, binary var
        self._C_b = {k: dict() for k in self._all_models}
        # C coefficients for bilinear terms, for the binary var
        self._C_b_q = {k: dict() for k in self._all_models}
        # C coefficients for linear terms, integer var
        self._C_i = {k: dict() for k in self._all_models}
        # C coefficients for bilinear terms, for the integer var
        self._C_i_q = {k: dict() for k in self._all_models}
        # C coefficients for linear terms, fixed var in SubModel
        self._C_f = {k: dict() for k in self._all_models - {model.name}}
        # C coefficients for bilinear terms, fixed var in SubModel
        self._C_f_q = {k: dict() for k in self._all_models - {model.name}}
        # C coefficients for constant terms
        self._C_constant = {k: dict() for k in self._all_models}

        self._preprocess()

    def cost_vectors(self, block, var):
        if id(block) != id(block.root_block()) and id(var) in self._fixed_var_ids[block.name]:
            C = (self._C_f.get(block.name)).get(id(var))
            C_q = (self._C_f_q.get(block.name)).get(id(var))
        else:
            if var.is_continuous():
                C = (self._C_c.get(block.name)).get(id(var))
                C_q = (self._C_c_q.get(block.name)).get(id(var))
            if var.is_binary():
                C = (self._C_b.get(block.name)).get(id(var))
                C_q = (self._C_c_q.get(block.name)).get(id(var))
            if var.is_integer():
                C = (self._C_i.get(block.name)).get(id(var))
                C_q = (self._C_i_q.get(block.name)).get(id(var))

        C_constant = self._C_constant.get(block.name)

        return C, C_q, C_constant  # returns coo_matrix for selected rows

    def coef_matrices(self, block, var, sense=None):
        if id(block) != id(block.root_block()) and id(var) in self._fixed_var_ids[block.name]:
            A = (self._A_f.get(block.name)).get(id(var))
            A_q = (self._A_f_q.get(block.name)).get(id(var))
        else:
            if var.is_continuous():
                A = (self._A_c.get(block.name)).get(id(var))
                A_q = (self._A_c_q.get(block.name)).get(id(var))
            if var.is_binary():
                A = (self._A_b.get(block.name)).get(id(var))
                A_q = (self._A_c_q.get(block.name)).get(id(var))
            if var.is_integer():
                A = (self._A_i.get(block.name)).get(id(var))
                A_q = (self._A_i_q.get(block.name)).get(id(var))

        # return the full matrices if the sign on the constraint is not specified
        cons_sense_rhs = self._cons_sense_rhs[block.name]
        b = [rhs for (_cid, _sense), rhs in cons_sense_rhs.items()]
        sign = [_sense if _sense != 'g->l' else 'l' for (_cid, _sense), rhs in cons_sense_rhs.items()]
        if sense is None or not sense in ['e', 'l', 'g','g->l']:
            return A, A_q, sign, array(b)

        # grouping 'l' (<=) and those 'g->l' transformed to (<=) together
        if sense=='l':
            sense = ['l','g->l']
        else:
            sense=[sense]

        # return partial matrices if the sign on the constraint is specified
        indices = list()
        b = list()
        sign = list()
        for (_cid, _sense), rhs in cons_sense_rhs.items():
            if _sense in sense:
                idx = list(cons_sense_rhs.keys()).index((_cid, _sense))
                indices.append(idx)
                b.append(rhs)
                if _sense == 'g->l':
                    sign.append('l')
                else:
                    sign.append(_sense)

        if len(indices) == 0:
            return array(indices), coo_matrix((0, 0)), sign, array(b) # return empty coo_matrix for A, A_q

        cons_sense_rhs = self._cons_sense_rhs[block.name]
        if len(indices) == len(cons_sense_rhs):
            return A, A_q, sign, array(b)  # return the full coo_matrix

        return A[indices], A_q.todok()[indices, :].tocoo(), sign, array(b)  # returns coo_matrix for selected rows

    def _preprocess(self):
        # Preprocess the main Model (upper-level)
        A_c, A_c_q, A_b, A_b_q, A_i, A_i_q, A_f, A_f_q, cons_sense_rhs = \
            collect_bilevel_constraint_matrix_representation(self._model, \
                                                             self._all_vars, \
                                                             self._c_var_ids, \
                                                             self._b_var_ids, \
                                                             self._i_var_ids, \
                                                             [], \
                                                             standard_form = self._standard_form)
        self._A_c[self._model.name] = A_c
        self._A_c_q[self._model.name] = A_c_q
        self._A_b[self._model.name] = A_b
        self._A_b_q[self._model.name] = A_b_q
        self._A_i[self._model.name] = A_i
        self._A_i_q[self._model.name] = A_i_q
        self._cons_sense_rhs[self._model.name] = cons_sense_rhs

        C_c, C_c_q, C_b, C_b_q, C_i, C_i_q, C_f, C_f_q, C_constant = \
            collect_bilevel_objective_vector_representation(self._model, \
                                                            self._all_vars, \
                                                            self._c_var_ids, \
                                                            self._b_var_ids, \
                                                            self._i_var_ids, \
                                                            [], \
                                                            standard_form = self._standard_form)
        self._C_c[self._model.name] = C_c
        self._C_c_q[self._model.name] = C_c_q
        self._C_b[self._model.name] = C_b
        self._C_b_q[self._model.name] = C_b_q
        self._C_i[self._model.name] = C_i
        self._C_i_q[self._model.name] = C_i_q
        self._C_constant[self._model.name] = C_constant

        # Preprocess each SubModel (lower-level(s))
        for block in self._model.component_objects(SubModel):
            A_c, A_c_q, A_b, A_b_q, A_i, A_i_q, A_f, A_f_q, cons_sense_rhs = \
                collect_bilevel_constraint_matrix_representation(block, \
                                                                 self._all_vars, \
                                                                 self._c_var_ids - self._fixed_var_ids[block.name], \
                                                                 self._b_var_ids - self._fixed_var_ids[block.name], \
                                                                 self._i_var_ids - self._fixed_var_ids[block.name], \
                                                                 self._fixed_var_ids[block.name], \
                                                                 standard_form = self._standard_form)
            self._A_c[block.name] = A_c
            self._A_c_q[block.name] = A_c_q
            self._A_b[block.name] = A_b
            self._A_b_q[block.name] = A_b_q
            self._A_i[block.name] = A_i
            self._A_i_q[block.name] = A_i_q
            self._A_f[block.name] = A_f
            self._A_f_q[block.name] = A_f_q
            self._cons_sense_rhs[block.name] = cons_sense_rhs

            C_c, C_c_q, C_b, C_b_q, C_i, C_i_q, C_f, C_f_q, C_constant = \
                collect_bilevel_objective_vector_representation(block, \
                                                                self._all_vars, \
                                                                self._c_var_ids - self._fixed_var_ids[block.name], \
                                                                self._b_var_ids - self._fixed_var_ids[block.name], \
                                                                self._i_var_ids - self._fixed_var_ids[block.name], \
                                                                self._fixed_var_ids[block.name], \
                                                                standard_form = self._standard_form)
            self._C_c[block.name] = C_c
            self._C_c_q[block.name] = C_c_q
            self._C_b[block.name] = C_b
            self._C_b_q[block.name] = C_b_q
            self._C_i[block.name] = C_i
            self._C_i_q[block.name] = C_i_q
            self._C_f[block.name] = C_f
            self._C_f_q[block.name] = C_f_q
            self._C_constant[block.name] = C_constant
